"""optimal_test.py - tests for optimization based control

RMM, 17 Apr 2019 check the functionality for optimization based control.
RMM, 30 Dec 2020 convert to pytest
"""

import pytest
import warnings
import numpy as np
import scipy as sp
import math
import control as ct
import control.optimal as opt
import control.flatsys as flat
from control.tests.conftest import slycotonly
from numpy.lib import NumpyVersion


@pytest.mark.parametrize("method, npts", [
    ('shooting', 5),
    ('collocation', 20),
])
def test_continuous_lqr(method, npts):
    # Create a lightly dampled, second order system
    sys = ct.ss([[0, 1], [-0.1, -0.01]], [[0], [1]], [[1, 0]], 0)

    # Define cost functions
    Q = np.eye(sys.nstates)
    R = np.eye(sys.ninputs) * 10

    # Figure out the LQR solution (for terminal cost)
    K, S, E = ct.lqr(sys, Q, R)

    # Define the cost functions
    traj_cost = opt.quadratic_cost(sys, Q, R)
    term_cost = opt.quadratic_cost(sys, S, None)
    constraints = opt.input_range_constraint(
        sys, -np.ones(sys.ninputs), np.ones(sys.ninputs))

    # Define the initial condition, time horizon, and time points
    x0 = np.ones(sys.nstates)
    Tf = 10
    timepts = np.linspace(0, Tf, npts)

    res = opt.solve_ocp(
        sys, timepts, x0, traj_cost, constraints, terminal_cost=term_cost,
        trajectory_method=method
    )

    # Make sure the optimization was successful
    assert res.success

    # Make sure we were reasonable close to the optimal cost
    assert res.cost / (x0 @ S @ x0) < 1.2       # shouldn't be too far off


@pytest.mark.parametrize("method", ['shooting']) # TODO: add 'collocation'
def test_finite_horizon_simple(method):
    # Define a (discrete time) linear system with constraints
    # Source: https://www.mpt3.org/UI/RegulationProblem

    # LTI prediction model (discrete time)
    sys = ct.ss2io(ct.ss([[1, 1], [0, 1]], [[1], [0.5]], np.eye(2), 0, 1))

    # State and input constraints
    constraints = [
        sp.optimize.LinearConstraint(np.eye(3), [-5, -5, -1], [5, 5, 1]),
    ]

    # Quadratic state and input penalty
    Q = [[1, 0], [0, 1]]
    R = [[1]]
    cost = opt.quadratic_cost(sys, Q, R)

    # Set up the optimal control problem
    time = np.arange(0, 5, 1)
    x0 = [4, 0]

    # Retrieve the full open-loop predictions
    res = opt.solve_ocp(
        sys, time, x0, cost, constraints, squeeze=True,
        trajectory_method=method,
        terminal_cost=cost)     # include to match MPT3 formulation
    t, u_openloop = res.time, res.inputs
    np.testing.assert_almost_equal(
        u_openloop, [-1, -1, 0.1393, 0.3361, -5.204e-16], decimal=4)

    # Make sure the final cost is correct
    assert math.isclose(res.cost, 32.4898, rel_tol=1e-5)

    # Convert controller to an explicit form (not implemented yet)
    # mpc_explicit = opt.explicit_mpc();

    # Test explicit controller
    # u_explicit = mpc_explicit(x0)
    # np.testing.assert_array_almost_equal(u_openloop, u_explicit)


#
# Compare to LQR solution
#
# The next unit test is intended to confirm that a finite horizon
# optimal control problem with terminal cost set to LQR "cost to go"
# gives the same answer as LQR.
#
@slycotonly
def test_discrete_lqr():
    # oscillator model defined in 2D
    # Source: https://www.mpt3.org/UI/RegulationProblem
    A = [[0.5403, -0.8415], [0.8415, 0.5403]]
    B = [[-0.4597], [0.8415]]
    C = [[1, 0]]
    D = [[0]]

    # Linear discrete-time model with sample time 1
    sys = ct.ss2io(ct.ss(A, B, C, D, 1))

    # Include weights on states/inputs
    Q = np.eye(2)
    R = 1
    K, S, E = ct.dlqr(A, B, Q, R)

    # Compute the integral and terminal cost
    integral_cost = opt.quadratic_cost(sys, Q, R)
    terminal_cost = opt.quadratic_cost(sys, S, None)

    # Solve the LQR problem
    lqr_sys = ct.ss2io(ct.ss(A - B @ K, B, C, D, 1))

    # Generate a simulation of the LQR controller
    time = np.arange(0, 5, 1)
    x0 = np.array([1, 1])
    _, _, lqr_x = ct.input_output_response(
        lqr_sys, time, 0, x0, return_x=True)

    # Use LQR input as initial guess to avoid convergence/precision issues
    lqr_u = np.array(-K @ lqr_x[0:time.size]) # convert from matrix

    # Formulate the optimal control problem and compute optimal trajectory
    optctrl = opt.OptimalControlProblem(
        sys, time, integral_cost, terminal_cost=terminal_cost,
        initial_guess=lqr_u)
    res1 = optctrl.compute_trajectory(x0, return_states=True)

    # Compare to make sure results are the same
    np.testing.assert_almost_equal(res1.inputs, lqr_u[0])
    np.testing.assert_almost_equal(res1.states, lqr_x)

    # Add state and input constraints
    trajectory_constraints = [
        sp.optimize.LinearConstraint(np.eye(3), [-5, -5, -.5], [5, 5, 0.5]),
    ]

    # Re-solve
    res2 = opt.solve_ocp(
        sys, time, x0, integral_cost, trajectory_constraints,
        terminal_cost=terminal_cost, initial_guess=lqr_u)

    # Make sure we got a different solution
    assert np.any(np.abs(res1.inputs - res2.inputs) > 0.1)


def test_mpc_iosystem_aircraft():
    # model of an aircraft discretized with 0.2s sampling time
    # Source: https://www.mpt3.org/UI/RegulationProblem
    A = [[0.99, 0.01, 0.18, -0.09,   0],
         [   0, 0.94,    0,  0.29,   0],
         [   0, 0.14, 0.81,  -0.9,   0],
         [   0, -0.2,    0,  0.95,   0],
         [   0, 0.09,    0,     0, 0.9]]
    B = [[ 0.01, -0.02],
         [-0.14,     0],
         [ 0.05,  -0.2],
         [ 0.02,     0],
         [-0.01, 0]]
    C = [[0, 1, 0, 0, -1],
         [0, 0, 1, 0,  0],
         [0, 0, 0, 1,  0],
         [1, 0, 0, 0,  0]]
    model = ct.ss2io(ct.ss(A, B, C, 0, 0.2))

    # For the simulation we need the full state output
    sys = ct.ss2io(ct.ss(A, B, np.eye(5), 0, 0.2))

    # compute the steady state values for a particular value of the input
    ud = np.array([0.8, -0.3])
    xd = np.linalg.inv(np.eye(5) - A) @ B @ ud
    yd = C @ xd

    # provide constraints on the system signals
    constraints = [opt.input_range_constraint(sys, [-5, -6], [5, 6])]

    # provide penalties on the system signals
    Q = model.C.transpose() @ np.diag([10, 10, 10, 10]) @ model.C
    R = np.diag([3, 2])
    cost = opt.quadratic_cost(model, Q, R, x0=xd, u0=ud)

    # online MPC controller object is constructed with a horizon 6
    ctrl = opt.create_mpc_iosystem(
        model, np.arange(0, 6) * 0.2, cost, constraints)

    # Define an I/O system implementing model predictive control
    loop = ct.feedback(sys, ctrl, 1)

    # Choose a nearby initial condition to speed up computation
    X0 = np.hstack([xd, np.kron(ud, np.ones(6))]) * 0.99

    Nsim = 12
    tout, xout = ct.input_output_response(
        loop, np.arange(0, Nsim) * 0.2, 0, X0)

    # Make sure the system converged to the desired state
    np.testing.assert_allclose(
        xout[0:sys.nstates, -1], xd, atol=0.1, rtol=0.01)


def test_mpc_iosystem_continuous():
    # Create a random state space system
    sys = ct.rss(2, 1, 1)
    T, _ = ct.step_response(sys)

    # provide penalties on the system signals
    Q = np.eye(sys.nstates)
    R = np.eye(sys.ninputs)
    cost = opt.quadratic_cost(sys, Q, R)

    # Continuous time MPC controller not implemented
    with pytest.raises(NotImplementedError):
        ctrl = opt.create_mpc_iosystem(sys, T, cost)


# Test various constraint combinations; need to use a somewhat convoluted
# parametrization due to the need to define sys instead the test function
@pytest.mark.parametrize("constraint_list", [
    [(sp.optimize.LinearConstraint, np.eye(3), [-5, -5, -1], [5, 5, 1],)],
    [(opt.state_range_constraint, [-5, -5], [5, 5]),
      (opt.input_range_constraint, [-1], [1])],
    [(opt.state_range_constraint, [-5, -5], [5, 5]),
      (opt.input_poly_constraint, np.array([[1], [-1]]), [1, 1])],
    [(opt.state_poly_constraint,
      np.array([[1, 0], [0, 1], [-1, 0], [0, -1]]), [5, 5, 5, 5]),
     (opt.input_poly_constraint, np.array([[1], [-1]]), [1, 1])],
    [(opt.output_range_constraint, [-5, -5], [5, 5]),
      (opt.input_poly_constraint, np.array([[1], [-1]]), [1, 1])],
    [(opt.output_poly_constraint,
      np.array([[1, 0], [0, 1], [-1, 0], [0, -1]]), [5, 5, 5, 5]),
     (opt.input_poly_constraint, np.array([[1], [-1]]), [1, 1])],
    [(sp.optimize.NonlinearConstraint,
      lambda x, u: np.array([x[0], x[1], u[0]]), [-5, -5, -1], [5, 5, 1])],
])
def test_constraint_specification(constraint_list):
    sys = ct.ss2io(ct.ss([[1, 1], [0, 1]], [[1], [0.5]], np.eye(2), 0, 1))

    """Test out different forms of constraints on a simple problem"""
    # Parse out the constraint
    constraints = []
    for constraint_setup in constraint_list:
        if constraint_setup[0] in \
           (sp.optimize.LinearConstraint, sp.optimize.NonlinearConstraint):
            # No processing required
            constraints.append(constraint_setup)
        else:
            # Call the function in the first argument to set up the constraint
            constraints.append(constraint_setup[0](sys, *constraint_setup[1:]))

    # Quadratic state and input penalty
    Q = [[1, 0], [0, 1]]
    R = [[1]]
    cost = opt.quadratic_cost(sys, Q, R)

    # Create a model predictive controller system
    time = np.arange(0, 5, 1)
    optctrl = opt.OptimalControlProblem(
        sys, time, cost, constraints,
        terminal_cost=cost)     # include to match MPT3 formulation

    # Compute optimal control and compare against MPT3 solution
    x0 = [4, 0]
    res = optctrl.compute_trajectory(x0, squeeze=True)
    t, u_openloop = res.time, res.inputs
    np.testing.assert_almost_equal(
        u_openloop, [-1, -1, 0.1393, 0.3361, -5.204e-16], decimal=3)


@pytest.mark.parametrize("sys_args", [
    pytest.param(
        ([[1, 0], [0, 1]], np.eye(2), np.eye(2), 0, True),
        id = "discrete, no timebase"),
    pytest.param(
        ([[1, 0], [0, 1]], np.eye(2), np.eye(2), 0, 1),
        id = "discrete, dt=1"),
    pytest.param(
        (np.zeros((2, 2)), np.eye(2), np.eye(2), 0),
        id = "continuous"),
])
def test_terminal_constraints(sys_args):
    """Test out the ability to handle terminal constraints"""
    # Create the system
    sys = ct.ss2io(ct.ss(*sys_args))

    # Shortest path to a point is a line
    Q = np.zeros((2, 2))
    R = np.eye(2)
    cost = opt.quadratic_cost(sys, Q, R)

    # Set up the terminal constraint to be the origin
    final_point = [opt.state_range_constraint(sys, [0, 0], [0, 0])]

    # Create the optimal control problem
    time = np.arange(0, 3, 1)
    optctrl = opt.OptimalControlProblem(
        sys, time, cost, terminal_constraints=final_point)

    # Find a path to the origin
    x0 = np.array([4, 3])
    res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)
    t, u1, x1 = res.time, res.inputs, res.states

    # Bug prior to SciPy 1.6 will result in incorrect results
    if NumpyVersion(sp.__version__) < '1.6.0':
        pytest.xfail("SciPy 1.6 or higher required")

    np.testing.assert_almost_equal(x1[:,-1], 0, decimal=4)

    # Make sure it is a straight line
    Tf = time[-1]
    if ct.isctime(sys):
        # Continuous time is not that accurate on the input, so just skip test
        pass
    else:
        # Final point doesn't affect cost => don't need to test
        np.testing.assert_almost_equal(
            u1[:, 0:-1],
            np.kron((-x0/Tf).reshape((2, 1)), np.ones(time.shape))[:, 0:-1])
    np.testing.assert_allclose(
        x1, np.kron(x0.reshape((2, 1)), time[::-1]/Tf), atol=0.1, rtol=0.01)

    # Re-run using initial guess = optional and make sure nothing changes
    res = optctrl.compute_trajectory(x0, initial_guess=u1)
    np.testing.assert_almost_equal(res.inputs, u1, decimal=2)
    np.testing.assert_almost_equal(res.states, x1, decimal=4)

    # Re-run using a basis function and see if we get the same answer
    res = opt.solve_ocp(
        sys, time, x0, cost, terminal_constraints=final_point,
        basis=flat.BezierFamily(8, Tf))

    # Final point doesn't affect cost => don't need to test
    np.testing.assert_almost_equal(
        res.inputs[:, :-1], u1[:, :-1], decimal=2)

    # Impose some cost on the state, which should change the path
    Q = np.eye(2)
    R = np.eye(2) * 0.1
    cost = opt.quadratic_cost(sys, Q, R)
    optctrl = opt.OptimalControlProblem(
        sys, time, cost, terminal_constraints=final_point)

    # Turn off warning messages, since we sometimes don't get convergence
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="unable to solve", category=UserWarning)
        # Find a path to the origin
        res = optctrl.compute_trajectory(
            x0, squeeze=True, return_x=True, initial_guess=u1)
        t, u2, x2 = res.time, res.inputs, res.states

        # Not all configurations are able to converge (?)
        if res.success:
            np.testing.assert_almost_equal(x2[:,-1], 0, decimal=5)

            # Make sure that it is *not* a straight line path
            assert np.any(np.abs(x2 - x1) > 0.1)
            assert np.any(np.abs(u2) > 1)       # Make sure next test is useful

        # Add some bounds on the inputs
        constraints = [opt.input_range_constraint(sys, [-1, -1], [1, 1])]
        optctrl = opt.OptimalControlProblem(
            sys, time, cost, constraints, terminal_constraints=final_point)
        res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)
        t, u3, x3 = res.time, res.inputs, res.states

        # Check the answers only if we converged
        if res.success:
            np.testing.assert_almost_equal(x3[:,-1], 0, decimal=4)

            # Make sure we got a new path and didn't violate the constraints
            assert np.any(np.abs(x3 - x1) > 0.1)
            np.testing.assert_array_less(np.abs(u3), 1 + 1e-6)

    # Make sure that infeasible problems are handled sensibly
    x0 = np.array([10, 3])
    with pytest.warns(UserWarning, match="unable to solve"):
        res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)
        assert not res.success


def test_optimal_logging(capsys):
    """Test logging functions (mainly for code coverage)"""
    sys = ct.ss2io(ct.ss(np.eye(2), np.eye(2), np.eye(2), 0, 1))

    # Set up the optimal control problem
    cost = opt.quadratic_cost(sys, 1, 1)
    state_constraint = opt.state_range_constraint(
        sys, [-np.inf, 1], [10, 1])
    input_constraint = opt.input_range_constraint(sys, [-100, -100], [100, 100])
    time = np.arange(0, 3, 1)
    x0 = [-1, 1]

    # Solve it, with logging turned on (with warning due to mixed constraints)
    with pytest.warns(sp.optimize.OptimizeWarning,
                        match="Equality and inequality .* same element"):
        res = opt.solve_ocp(
            sys, time, x0, cost, input_constraint, terminal_cost=cost,
            terminal_constraints=state_constraint, log=True)

    # Make sure the output has info available only with logging turned on
    captured = capsys.readouterr()
    assert captured.out.find("process time") != -1


@pytest.mark.parametrize("fun, args, exception, match", [
    [opt.quadratic_cost, (np.zeros((2, 3)), np.eye(2)), ValueError,
     "Q matrix is the wrong shape"],
    [opt.quadratic_cost, (np.eye(2), np.eye(2, 3)), ValueError,
     "R matrix is the wrong shape"],
    [opt.input_poly_constraint, (np.zeros((2, 3)), [0, 0]), ValueError,
     "polytope matrix must match number of inputs"],
    [opt.output_poly_constraint, (np.zeros((2, 3)), [0, 0]), ValueError,
     "polytope matrix must match number of outputs"],
    [opt.state_poly_constraint, (np.zeros((2, 3)), [0, 0]), ValueError,
     "polytope matrix must match number of states"],
    [opt.input_poly_constraint, (np.zeros((2, 2)), [0, 0, 0]), ValueError,
     "number of bounds must match number of constraints"],
    [opt.output_poly_constraint, (np.zeros((2, 2)), [0, 0, 0]), ValueError,
     "number of bounds must match number of constraints"],
    [opt.state_poly_constraint, (np.zeros((2, 2)), [0, 0, 0]), ValueError,
     "number of bounds must match number of constraints"],
    [opt.input_poly_constraint, (np.zeros((2, 2)), [[0, 0, 0]]), ValueError,
     "number of bounds must match number of constraints"],
    [opt.output_poly_constraint, (np.zeros((2, 2)), [[0, 0, 0]]), ValueError,
     "number of bounds must match number of constraints"],
    [opt.state_poly_constraint, (np.zeros((2, 2)), 0), ValueError,
     "number of bounds must match number of constraints"],
    [opt.input_range_constraint, ([1, 2, 3], [0, 0]), ValueError,
     "input bounds must match"],
    [opt.output_range_constraint, ([2, 3], [0, 0, 0]), ValueError,
     "output bounds must match"],
    [opt.state_range_constraint, ([1, 2, 3], [0, 0, 0]), ValueError,
     "state bounds must match"],
])
def test_constraint_constructor_errors(fun, args, exception, match):
    """Test various error conditions for constraint constructors"""
    sys = ct.ss2io(ct.rss(2, 2, 2))
    with pytest.raises(exception, match=match):
        fun(sys, *args)


def test_ocp_argument_errors():
    sys = ct.ss2io(ct.ss([[1, 1], [0, 1]], [[1], [0.5]], np.eye(2), 0, 1))

    # State and input constraints
    constraints = [
        sp.optimize.LinearConstraint(np.eye(3), [-5, -5, -1], [5, 5, 1]),
    ]

    # Quadratic state and input penalty
    Q = [[1, 0], [0, 1]]
    R = [[1]]
    cost = opt.quadratic_cost(sys, Q, R)

    # Set up the optimal control problem
    time = np.arange(0, 5, 1)
    x0 = [4, 0]

    # Trajectory constraints not in the right form
    with pytest.raises(TypeError, match="constraints must be a list"):
        res = opt.solve_ocp(sys, time, x0, cost, np.eye(2))

    # Terminal constraints not in the right form
    with pytest.raises(TypeError, match="constraints must be a list"):
        res = opt.solve_ocp(
            sys, time, x0, cost, constraints, terminal_constraints=np.eye(2))

    # Initial guess in the wrong shape
    with pytest.raises(ValueError, match="initial guess is the wrong shape"):
        res = opt.solve_ocp(
            sys, time, x0, cost, constraints, initial_guess=np.zeros((4,1,1)))

    # Unrecognized arguments
    with pytest.raises(TypeError, match="unrecognized keyword"):
        res = opt.solve_ocp(
            sys, time, x0, cost, constraints, terminal_constraint=None)

    # Unrecognized trajectory constraint type
    constraints = [(None, np.eye(3), [0, 0, 0], [0, 0, 0])]
    with pytest.raises(TypeError, match="unknown trajectory constraint type"):
        res = opt.solve_ocp(
            sys, time, x0, cost, trajectory_constraints=constraints)

    # Unrecognized terminal constraint type
    with pytest.raises(TypeError, match="unknown terminal constraint type"):
        res = opt.solve_ocp(
            sys, time, x0, cost, terminal_constraints=constraints)

    # Discrete time system checks: solve_ivp keywords not allowed
    sys = ct.rss(2, 1, 1, dt=True)
    with pytest.raises(TypeError, match="solve_ivp method, kwargs not allowed"):
        res = opt.solve_ocp(
            sys, time, x0, cost, solve_ivp_method='LSODA')
    with pytest.raises(TypeError, match="solve_ivp method, kwargs not allowed"):
        res = opt.solve_ocp(
            sys, time, x0, cost, solve_ivp_kwargs={'eps': 0.1})


@pytest.mark.parametrize("basis", [
    flat.PolyFamily(4), flat.PolyFamily(6),
    flat.BezierFamily(4), flat.BSplineFamily([0, 4, 8], 6)
    ])
def test_optimal_basis_simple(basis):
    sys = ct.ss([[1, 1], [0, 1]], [[1], [0.5]], np.eye(2), 0, 1)

    # State and input constraints
    constraints = [
        sp.optimize.LinearConstraint(np.eye(3), [-5, -5, -1], [5, 5, 1]),
    ]

    # Quadratic state and input penalty
    Q = [[1, 0], [0, 1]]
    R = [[1]]
    cost = opt.quadratic_cost(sys, Q, R)

    # Set up the optimal control problem
    Tf = 5
    time = np.arange(0, Tf, 1)
    x0 = [4, 0]

    # Basic optimal control problem
    res1 = opt.solve_ocp(
        sys, time, x0, cost, constraints,
        terminal_cost=cost, basis=basis, return_x=True)
    assert res1.success

    # Make sure the constraints were satisfied
    np.testing.assert_array_less(np.abs(res1.states[0]), 5 + 1e-6)
    np.testing.assert_array_less(np.abs(res1.states[1]), 5 + 1e-6)
    np.testing.assert_array_less(np.abs(res1.inputs[0]), 1 + 1e-6)

    # Pass an initial guess and rerun
    res2 = opt.solve_ocp(
        sys, time, x0, cost, constraints, initial_guess=0.99*res1.inputs,
        terminal_cost=cost, basis=basis, return_x=True)
    assert res2.success
    np.testing.assert_allclose(res2.inputs, res1.inputs, atol=0.01, rtol=0.01)

    # Run with logging turned on for code coverage
    res3 = opt.solve_ocp(
        sys, time, x0, cost, constraints, terminal_cost=cost,
        basis=basis, return_x=True, log=True)
    assert res3.success
    np.testing.assert_almost_equal(res3.inputs, res1.inputs, decimal=3)


def test_equality_constraints():
    """Test out the ability to handle equality constraints"""
    # Create the system (double integrator, continuous time)
    sys = ct.ss2io(ct.ss(np.zeros((2, 2)), np.eye(2), np.eye(2), 0))

    # Shortest path to a point is a line
    Q = np.zeros((2, 2))
    R = np.eye(2)
    cost = opt.quadratic_cost(sys, Q, R)

    # Set up the terminal constraint to be the origin
    final_point = [opt.state_range_constraint(sys, [0, 0], [0, 0])]

    # Create the optimal control problem
    time = np.arange(0, 3, 1)
    optctrl = opt.OptimalControlProblem(
        sys, time, cost, terminal_constraints=final_point)

    # Find a path to the origin
    x0 = np.array([4, 3])
    res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)
    t, u1, x1 = res.time, res.inputs, res.states

    # Bug prior to SciPy 1.6 will result in incorrect results
    if NumpyVersion(sp.__version__) < '1.6.0':
        pytest.xfail("SciPy 1.6 or higher required")

    np.testing.assert_almost_equal(x1[:,-1], 0, decimal=4)

    # Set up terminal constraints as a nonlinear constraint
    def final_point_eval(x, u):
        return x
    final_point = [
        sp.optimize.NonlinearConstraint(final_point_eval, [0, 0], [0, 0])]

    optctrl = opt.OptimalControlProblem(
        sys, time, cost, terminal_constraints=final_point)

    # Find a path to the origin
    x0 = np.array([4, 3])
    res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)
    t, u2, x2 = res.time, res.inputs, res.states
    np.testing.assert_almost_equal(x2[:,-1], 0, decimal=4)
    np.testing.assert_almost_equal(u1, u2)
    np.testing.assert_almost_equal(x1, x2)

    # Try passing and unknown constraint type
    final_point = [(None, final_point_eval, [0, 0], [0, 0])]
    with pytest.raises(TypeError, match="unknown terminal constraint type"):
        optctrl = opt.OptimalControlProblem(
            sys, time, cost, terminal_constraints=final_point)
        res = optctrl.compute_trajectory(x0, squeeze=True, return_x=True)


@pytest.mark.parametrize(
    "method, npts, initial_guess, fail", [
        ('shooting', 3, None, 'xfail'),         # doesn't converge
        ('shooting', 3, 'zero', 'xfail'),       # doesn't converge
        ('shooting', 3, 'u0', None),            # github issue #782
        ('shooting', 3, 'input', 'endpoint'),   # doesn't converge to optimal
        ('shooting', 5, 'input', 'endpoint'),   # doesn't converge to optimal
        ('collocation', 3, 'u0', 'endpoint'),   # doesn't converge to optimal
        ('collocation', 5, 'u0', 'endpoint'),
        ('collocation', 5, 'input', 'openloop'),# open loop sim fails
        ('collocation', 10, 'input', None),
        ('collocation', 10, 'u0', None),        # from documentation
        ('collocation', 10, 'state', None),
        ('collocation', 20, 'state', None),
    ])
def test_optimal_doc(method, npts, initial_guess, fail):
    """Test optimal control problem from documentation"""
    def vehicle_update(t, x, u, params):
        # Get the parameters for the model
        l = params.get('wheelbase', 3.)         # vehicle wheelbase
        phimax = params.get('maxsteer', 0.5)    # max steering angle (rad)

        # Saturate the steering input
        phi = np.clip(u[1], -phimax, phimax)

        # Return the derivative of the state
        return np.array([
            np.cos(x[2]) * u[0],            # xdot = cos(theta) v
            np.sin(x[2]) * u[0],            # ydot = sin(theta) v
            (u[0] / l) * np.tan(phi)        # thdot = v/l tan(phi)
        ])

    def vehicle_output(t, x, u, params):
        return x                            # return x, y, theta (full state)

    # Define the vehicle steering dynamics as an input/output system
    vehicle = ct.NonlinearIOSystem(
        vehicle_update, vehicle_output, states=3, name='vehicle',
        inputs=('v', 'phi'), outputs=('x', 'y', 'theta'))

    # Define the initial and final points and time interval
    x0 = np.array([0., -2., 0.]); u0 = np.array([10., 0.])
    xf = np.array([100., 2., 0.]); uf = np.array([10., 0.])
    Tf = 10

    # Define the cost functions
    Q = np.diag([0, 0, 0.1])          # don't turn too sharply
    R = np.diag([1, 1])               # keep inputs small
    P = np.diag([1000, 1000, 1000])   # get close to final point
    traj_cost = opt.quadratic_cost(vehicle, Q, R, x0=xf, u0=uf)
    term_cost = opt.quadratic_cost(vehicle, P, 0, x0=xf)

    # Define the constraints
    constraints = [ opt.input_range_constraint(vehicle, [8, -0.1], [12, 0.1]) ]

    # Define an initial guess at the trajectory
    timepts = np.linspace(0, Tf, npts, endpoint=True)
    if initial_guess == 'zero':
        initial_guess = 0

    elif initial_guess == 'u0':
        initial_guess = u0

    elif initial_guess == 'input':
        # Velocity = constant that gets us from start to end
        initial_guess = np.zeros((vehicle.ninputs, timepts.size))
        initial_guess[0, :] = (xf[0] - x0[0]) / Tf

        # Steering = rate required to turn to proper slope in first segment
        straight_seg_length = timepts[-2] - timepts[1]
        curved_seg_length = (Tf - straight_seg_length)/2
        approximate_angle = math.atan2(xf[1] - x0[1], xf[0] - x0[0])
        initial_guess[1, 0] = approximate_angle / (timepts[1] - timepts[0])
        initial_guess[1, -1] = -approximate_angle / (timepts[-1] - timepts[-2])

    elif initial_guess == 'state':
        input_guess = np.outer(u0, np.ones((1, npts)))
        state_guess = np.array([
            x0 + (xf - x0) * time/Tf for time in timepts]).transpose()
        initial_guess = (state_guess, input_guess)

    # Solve the optimal control problem
    result = opt.solve_ocp(
        vehicle, timepts, x0, traj_cost, constraints,
        terminal_cost=term_cost, initial_guess=initial_guess,
        trajectory_method=method,
        # minimize_method='COBYLA', # SLSQP',
    )

    if fail == 'xfail':
        assert not result.success
        pytest.xfail("optimization fails to converge")
    elif fail == 'precision':
        assert result.status == 2
        pytest.xfail("optimization precision not achieved")
    else:
        # Make sure the optimization was successful
        assert result.success

        # Make sure we started and stopped at the right spot
        if fail == 'endpoint':
            assert not np.allclose(result.states[:, -1], xf, rtol=1e-4)
            pytest.xfail("optimization does not converge to endpoint")
        else:
            np.testing.assert_almost_equal(result.states[:, 0], x0, decimal=4)
            np.testing.assert_almost_equal(result.states[:, -1], xf, decimal=2)

            # Simulate the trajectory to make sure it looks OK
            resp = ct.input_output_response(
                vehicle, timepts, result.inputs, x0,
                t_eval=np.linspace(0, Tf, 10))
            t, y = resp
            if fail == 'openloop':
                with pytest.raises(AssertionError):
                    np.testing.assert_almost_equal(y[:,-1], xf, decimal=1)
            else:
                np.testing.assert_almost_equal(y[:,-1], xf, decimal=1)



# ctrlutil.py - control system utility functions
#
# Author: Richard M. Murray
# Date: 24 May 09
#
# These are some basic utility functions that are used in the control
# systems library and that didn't naturally fit anyplace else.
#
# Copyright (c) 2009 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id$

# Packages that we need access to
from . import lti
import numpy as np
import math

__all__ = ['unwrap', 'issys', 'db2mag', 'mag2db']

# Utility function to unwrap an angle measurement
def unwrap(angle, period=2*math.pi):
    """Unwrap a phase angle to give a continuous curve

    Parameters
    ----------
    angle : array_like
        Array of angles to be unwrapped
    period : float, optional
        Period (defaults to `2*pi`)

    Returns
    -------
    angle_out : array_like
        Output array, with jumps of period/2 eliminated

    Examples
    --------
    >>> import numpy as np
    >>> theta = [5.74, 5.97, 6.19, 0.13, 0.35, 0.57]
    >>> unwrap(theta, period=2 * np.pi)
    [5.74, 5.97, 6.19, 6.413185307179586, 6.633185307179586, 6.8531853071795865]

    """
    dangle = np.diff(angle)
    dangle_desired = (dangle + period/2.) % period - period/2.
    correction = np.cumsum(dangle_desired - dangle)
    angle[1:] += correction
    return angle

def issys(obj):
    """Return True if an object is a system, otherwise False"""
    return isinstance(obj, lti.LTI)

def db2mag(db):
    """Convert a gain in decibels (dB) to a magnitude

    If A is magnitude,

        db = 20 * log10(A)

    Parameters
    ----------
    db : float or ndarray
        input value or array of values, given in decibels

    Returns
    -------
    mag : float or ndarray
        corresponding magnitudes

    """
    return 10. ** (db / 20.)

def mag2db(mag):
    """Convert a magnitude to decibels (dB)

    If A is magnitude,

        db = 20 * log10(A)

    Parameters
    ----------
    mag : float or ndarray
        input magnitude or array of magnitudes

    Returns
    -------
    db : float or ndarray
        corresponding values in decibels

    """
    return 20. * np.log10(mag)



"""iosys_test.py - test input/output system operations

RMM, 17 Apr 2019

This test suite checks to make sure that basic input/output class
operations are working.  It doesn't do exhaustive testing of
operations on input/output systems.  Separate unit tests should be
created for that purpose.
"""

import re
import warnings
import pytest

import numpy as np
from math import sqrt

import control as ct
from control import iosys as ios
from control.tests.conftest import matrixfilter


class TestIOSys:

    @pytest.fixture
    def tsys(self):
        class TSys:
            pass
        T = TSys()
        """Return some test systems"""
        # Create a single input/single output linear system
        T.siso_linsys = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0], [1]], [[1, 0]], [[0]])

        # Create a multi input/multi output linear system
        T.mimo_linsys1 = ct.StateSpace(
            [[-1, 1], [0, -2]], [[1, 0], [0, 1]],
            [[1, 0], [0, 1]], np.zeros((2, 2)))

        # Create a multi input/multi output linear system
        T.mimo_linsys2 = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0, 1], [1, 0]],
            [[1, 0], [0, 1]], np.zeros((2, 2)))

        # Create a static gain linear system
        T.staticgain = ct.StateSpace([], [], [], 1)

        # Create simulation parameters
        T.T = np.linspace(0, 10, 100)
        T.U = np.sin(T.T)
        T.X0 = [0, 0]

        return T

    def test_linear_iosys(self, tsys):
        # Create an input/output system from the linear system
        linsys = tsys.siso_linsys
        iosys = ios.LinearIOSystem(linsys).copy()

        # Make sure that the right hand side matches linear system
        for x, u in (([0, 0], 0), ([1, 0], 0), ([0, 1], 0), ([0, 0], 1)):
            np.testing.assert_array_almost_equal(
                np.reshape(iosys._rhs(0, x, u), (-1, 1)),
                linsys.A @ np.reshape(x, (-1, 1)) + linsys.B * u)

        # Make sure that simulations also line up
        T, U, X0 = tsys.T, tsys.U, tsys.X0
        lti_t, lti_y = ct.forced_response(linsys, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y, atol=0.002, rtol=0.)

        # Make sure that a static linear system has dt=None
        # and otherwise dt is as specified
        assert ios.LinearIOSystem(tsys.staticgain).dt is None
        assert ios.LinearIOSystem(tsys.staticgain, dt=.1).dt == .1

    def test_tf2io(self, tsys):
        # Create a transfer function from the state space system
        linsys = tsys.siso_linsys
        tfsys = ct.ss2tf(linsys)
        iosys = ct.tf2io(tfsys)

        # Verify correctness via simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0
        lti_t, lti_y = ct.forced_response(linsys, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y, atol=0.002, rtol=0.)

        # Make sure that non-proper transfer functions generate an error
        tfsys = ct.tf('s')
        with pytest.raises(ValueError):
            iosys=ct.tf2io(tfsys)

    def test_ss2io(self, tsys):
        # Create an input/output system from the linear system
        linsys = tsys.siso_linsys
        iosys = ct.ss2io(linsys)
        np.testing.assert_allclose(linsys.A, iosys.A)
        np.testing.assert_allclose(linsys.B, iosys.B)
        np.testing.assert_allclose(linsys.C, iosys.C)
        np.testing.assert_allclose(linsys.D, iosys.D)

        # Try adding names to things
        iosys_named = ct.ss2io(linsys, inputs='u', outputs='y',
                               states=['x1', 'x2'], name='iosys_named')
        assert iosys_named.find_input('u') == 0
        assert iosys_named.find_input('x') is None
        assert iosys_named.find_output('y') == 0
        assert iosys_named.find_output('u') is None
        assert iosys_named.find_state('x0') is None
        assert iosys_named.find_state('x1') == 0
        assert iosys_named.find_state('x2') == 1
        np.testing.assert_allclose(linsys.A, iosys_named.A)
        np.testing.assert_allclose(linsys.B, iosys_named.B)
        np.testing.assert_allclose(linsys.C, iosys_named.C)
        np.testing.assert_allclose(linsys.D, iosys_named.D)

    def test_iosys_unspecified(self, tsys):
        """System with unspecified inputs and outputs"""
        sys = ios.NonlinearIOSystem(secord_update, secord_output)
        np.testing.assert_raises(TypeError, sys.__mul__, sys)

    def test_iosys_print(self, tsys, capsys):
        """Make sure we can print various types of I/O systems"""
        # Send the output to /dev/null

        # Simple I/O system
        iosys = ct.ss2io(tsys.siso_linsys)
        print(iosys)

        # I/O system without ninputs, noutputs
        ios_unspecified = ios.NonlinearIOSystem(secord_update, secord_output)
        print(ios_unspecified)

        # I/O system with derived inputs and outputs
        ios_linearized = ios.linearize(ios_unspecified, [0, 0], [0])
        print(ios_linearized)

    @pytest.mark.parametrize("ss", [ios.NonlinearIOSystem, ct.ss])
    def test_nonlinear_iosys(self, tsys, ss):
        # Create a simple nonlinear I/O system
        nlsys = ios.NonlinearIOSystem(predprey)
        T = tsys.T

        # Start by simulating from an equilibrium point
        X0 = [0, 0]
        ios_t, ios_y = ios.input_output_response(nlsys, T, 0, X0)
        np.testing.assert_array_almost_equal(ios_y, np.zeros(np.shape(ios_y)))

        # Now simulate from a nonzero point
        X0 = [0.5, 0.5]
        ios_t, ios_y = ios.input_output_response(nlsys, T, 0, X0)

        #
        # Simulate a linear function as a nonlinear function and compare
        #
        # Create a single input/single output linear system
        linsys = tsys.siso_linsys

        # Create a nonlinear system with the same dynamics
        nlupd = lambda t, x, u, params: \
            np.reshape(linsys.A @ np.reshape(x, (-1, 1))
                       + linsys.B @ np.reshape(u, (-1, 1)),
                       (-1,))
        nlout = lambda t, x, u, params: \
            np.reshape(linsys.C @ np.reshape(x, (-1, 1))
                       + linsys.D @ np.reshape(u, (-1, 1)),
                       (-1,))
        nlsys = ios.NonlinearIOSystem(nlupd, nlout, inputs=1, outputs=1)

        # Make sure that simulations also line up
        T, U, X0 = tsys.T, tsys.U, tsys.X0
        lti_t, lti_y = ct.forced_response(linsys, T, U, X0)
        ios_t, ios_y = ios.input_output_response(nlsys, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y,atol=0.002,rtol=0.)

    @pytest.fixture
    def kincar(self):
        # Create a simple nonlinear system to check (kinematic car)
        def kincar_update(t, x, u, params):
            return np.array([np.cos(x[2]) * u[0], np.sin(x[2]) * u[0], u[1]])

        def kincar_output(t, x, u, params):
            return np.array([x[0], x[1]])

        return ios.NonlinearIOSystem(
            kincar_update, kincar_output,
            inputs = ['v', 'phi'],
            outputs = ['x', 'y'],
            states = ['x', 'y', 'theta'])

    def test_linearize(self, tsys, kincar):
        # Create a single input/single output linear system
        linsys = tsys.siso_linsys
        iosys = ios.LinearIOSystem(linsys)

        # Linearize it and make sure we get back what we started with
        linearized = iosys.linearize([0, 0], 0)
        np.testing.assert_array_almost_equal(linsys.A, linearized.A)
        np.testing.assert_array_almost_equal(linsys.B, linearized.B)
        np.testing.assert_array_almost_equal(linsys.C, linearized.C)
        np.testing.assert_array_almost_equal(linsys.D, linearized.D)

        # Create a simple nonlinear system to check (kinematic car)
        iosys = kincar
        linearized = iosys.linearize([0, 0, 0], [0, 0])
        np.testing.assert_array_almost_equal(linearized.A, np.zeros((3,3)))
        np.testing.assert_array_almost_equal(
            linearized.B, [[1, 0], [0, 0], [0, 1]])
        np.testing.assert_array_almost_equal(
            linearized.C, [[1, 0, 0], [0, 1, 0]])
        np.testing.assert_array_almost_equal(linearized.D, np.zeros((2,2)))

    @pytest.mark.usefixtures("editsdefaults")
    def test_linearize_named_signals(self, kincar):
        # Full form of the call
        linearized = kincar.linearize([0, 0, 0], [0, 0], copy_names=True,
                                      name='linearized')
        assert linearized.name == 'linearized'
        assert linearized.find_input('v') == 0
        assert linearized.find_input('phi') == 1
        assert linearized.find_output('x') == 0
        assert linearized.find_output('y') == 1
        assert linearized.find_state('x') == 0
        assert linearized.find_state('y') == 1
        assert linearized.find_state('theta') == 2

        # If we copy signal names w/out a system name, append '$linearized'
        linearized = kincar.linearize([0, 0, 0], [0, 0], copy_names=True)
        assert linearized.name == kincar.name + '$linearized'

        # If copy is False, signal names should not be copied
        lin_nocopy = kincar.linearize(0, 0, copy_names=False)
        assert lin_nocopy.find_input('v') is None
        assert lin_nocopy.find_output('x') is None
        assert lin_nocopy.find_state('x') is None

        # if signal names are provided, they should override those of kincar
        linearized_newnames = kincar.linearize([0, 0, 0], [0, 0],
            name='linearized',
            copy_names=True, inputs=['v2', 'phi2'], outputs=['x2','y2'])
        assert linearized_newnames.name == 'linearized'
        assert linearized_newnames.find_input('v2') == 0
        assert linearized_newnames.find_input('phi2') == 1
        assert linearized_newnames.find_input('v') is None
        assert linearized_newnames.find_input('phi') is None
        assert linearized_newnames.find_output('x2') == 0
        assert linearized_newnames.find_output('y2') == 1
        assert linearized_newnames.find_output('x') is None
        assert linearized_newnames.find_output('y') is None

        # Test legacy version as well
        ct.use_legacy_defaults('0.8.4')
        ct.config.use_numpy_matrix(False)       # np.matrix deprecated
        linearized = kincar.linearize([0, 0, 0], [0, 0], copy_names=True)
        assert linearized.name == kincar.name + '_linearized'

    def test_connect(self, tsys):
        # Define a couple of (linear) systems to interconnection
        linsys1 = tsys.siso_linsys
        iosys1 = ios.LinearIOSystem(linsys1, name='iosys1')
        linsys2 = tsys.siso_linsys
        iosys2 = ios.LinearIOSystem(linsys2, name='iosys2')

        # Connect systems in different ways and compare to StateSpace
        linsys_series = linsys2 * linsys1
        iosys_series = ios.InterconnectedSystem(
            [iosys1, iosys2],   # systems
            [[1, 0]],           # interconnection (series)
            0,                  # input = first system
            1                   # output = second system
        )

        # Run a simulation and compare to linear response
        T, U = tsys.T, tsys.U
        X0 = np.concatenate((tsys.X0, tsys.X0))
        ios_t, ios_y, ios_x = ios.input_output_response(
            iosys_series, T, U, X0, return_x=True)
        lti_t, lti_y = ct.forced_response(linsys_series, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y,atol=0.002,rtol=0.)

        # Connect systems with different timebases
        linsys2c = tsys.siso_linsys
        linsys2c.dt = 0         # Reset the timebase
        iosys2c = ios.LinearIOSystem(linsys2c)
        iosys_series = ios.InterconnectedSystem(
            [iosys1, iosys2c],   # systems
            [[1, 0]],          # interconnection (series)
            0,                  # input = first system
            1                   # output = second system
        )
        assert ct.isctime(iosys_series, strict=True)
        ios_t, ios_y, ios_x = ios.input_output_response(
            iosys_series, T, U, X0, return_x=True)
        lti_t, lti_y = ct.forced_response(linsys_series, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y,atol=0.002,rtol=0.)

        # Feedback interconnection
        linsys_feedback = ct.feedback(linsys1, linsys2)
        iosys_feedback = ios.InterconnectedSystem(
            [iosys1, iosys2],   # systems
            [[1, 0],            # input of sys2 = output of sys1
             [0, (1, 0, -1)]],  # input of sys1 = -output of sys2
            0,                  # input = first system
            0                   # output = first system
        )
        ios_t, ios_y, ios_x = ios.input_output_response(
            iosys_feedback, T, U, X0, return_x=True)
        lti_t, lti_y = ct.forced_response(linsys_feedback, T, U, X0)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y,atol=0.002,rtol=0.)

    @pytest.mark.parametrize(
        "connections, inplist, outlist",
        [pytest.param([[(1, 0), (0, 0, 1)]], [[(0, 0, 1)]], [[(1, 0, 1)]],
                      id="full, raw tuple"),
         pytest.param([[(1, 0), (0, 0, -1)]], [[(0, 0)]], [[(1, 0, -1)]],
                      id="full, raw tuple, canceling gains"),
         pytest.param([[(1, 0), (0, 0)]], [[(0, 0)]], [[(1, 0)]],
                      id="full, raw tuple, no gain"),
         pytest.param([[(1, 0), (0, 0)]], [(0, 0)], [(1, 0)],
                      id="full, raw tuple, no gain, no outer list"),
         pytest.param([['sys2.u[0]', 'sys1.y[0]']], ['sys1.u[0]'],
                      ['sys2.y[0]'], id="named, full"),
         pytest.param([['sys2.u[0]', '-sys1.y[0]']], ['sys1.u[0]'],
                      ['-sys2.y[0]'], id="named, full, caneling gains"),
         pytest.param([['sys2.u[0]', 'sys1.y[0]']], 'sys1.u[0]', 'sys2.y[0]',
                      id="named, full, no list"),
         pytest.param([['sys2.u[0]', ('sys1', 'y[0]')]], [(0, 0)], [(1,)],
                      id="mixed"),
         pytest.param([[1, 0]], 0, 1, id="minimal")])
    def test_connect_spec_variants(self, tsys, connections, inplist, outlist):
        # Define a couple of (linear) systems to interconnection
        linsys1 = tsys.siso_linsys
        iosys1 = ios.LinearIOSystem(linsys1, name="sys1")
        linsys2 = tsys.siso_linsys
        iosys2 = ios.LinearIOSystem(linsys2, name="sys2")

        # Simple series connection
        linsys_series = linsys2 * linsys1

        # Create a simulation run to compare against
        T, U = tsys.T, tsys.U
        X0 = np.concatenate((tsys.X0, tsys.X0))
        lti_t, lti_y, lti_x = ct.forced_response(
            linsys_series, T, U, X0, return_x=True)

        # Create the input/output system with different parameter variations
        iosys_series = ios.InterconnectedSystem(
            [iosys1, iosys2], connections, inplist, outlist)
        ios_t, ios_y, ios_x = ios.input_output_response(
            iosys_series, T, U, X0, return_x=True)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y, atol=0.002, rtol=0.)

    @pytest.mark.parametrize(
        "connections, inplist, outlist",
        [pytest.param([['sys2.u[0]', 'sys1.y[0]']],
                      [[('sys1', 'u[0]'), ('sys1', 'u[0]')]],
                      [('sys2', 'y[0]', 0.5)], id="duplicated input"),
         pytest.param([['sys2.u[0]', ('sys1', 'y[0]', 0.5)],
                       ['sys2.u[0]', ('sys1', 'y[0]', 0.5)]],
                      'sys1.u[0]', 'sys2.y[0]', id="duplicated connection"),
         pytest.param([['sys2.u[0]', 'sys1.y[0]']], 'sys1.u[0]',
                      [[('sys2', 'y[0]', 0.5), ('sys2', 'y[0]', 0.5)]],
                      id="duplicated output")])
    def test_connect_spec_warnings(self, tsys, connections, inplist, outlist):
        # Define a couple of (linear) systems to interconnection
        linsys1 = tsys.siso_linsys
        iosys1 = ios.LinearIOSystem(linsys1, name="sys1")
        linsys2 = tsys.siso_linsys
        iosys2 = ios.LinearIOSystem(linsys2, name="sys2")

        # Simple series connection
        linsys_series = linsys2 * linsys1

        # Create a simulation run to compare against
        T, U = tsys.T, tsys.U
        X0 = np.concatenate((tsys.X0, tsys.X0))
        lti_t, lti_y, lti_x = ct.forced_response(
            linsys_series, T, U, X0, return_x=True)

        # Set up multiple gainst and make sure a warning is generated
        with pytest.warns(UserWarning, match="multiple.*Combining"):
            iosys_series = ios.InterconnectedSystem(
                [iosys1, iosys2], connections, inplist, outlist)
        ios_t, ios_y, ios_x = ios.input_output_response(
            iosys_series, T, U, X0, return_x=True)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y, atol=0.002, rtol=0.)

    def test_static_nonlinearity(self, tsys):
        # Linear dynamical system
        linsys = tsys.siso_linsys
        ioslin = ios.LinearIOSystem(linsys)

        # Nonlinear saturation
        sat = lambda u: u if abs(u) < 1 else np.sign(u)
        sat_output = lambda t, x, u, params: sat(u)
        nlsat =  ios.NonlinearIOSystem(None, sat_output, inputs=1, outputs=1)

        # Set up parameters for simulation
        T, U, X0 = tsys.T, 2 * tsys.U, tsys.X0
        Usat = np.vectorize(sat)(U)

        # Make sure saturation works properly by comparing linear system with
        # saturated input to nonlinear system with saturation composition
        lti_t, lti_y, lti_x = ct.forced_response(
            linsys, T, Usat, X0, return_x=True)
        ios_t, ios_y, ios_x = ios.input_output_response(
            ioslin * nlsat, T, U, X0, return_x=True)
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_array_almost_equal(lti_y, ios_y, decimal=2)


    @pytest.mark.filterwarnings("ignore:Duplicate name::control.iosys")
    def test_algebraic_loop(self, tsys):
        # Create some linear and nonlinear systems to play with
        linsys = tsys.siso_linsys
        lnios = ios.LinearIOSystem(linsys)
        nlios =  ios.NonlinearIOSystem(None, \
            lambda t, x, u, params: u*u, inputs=1, outputs=1)
        nlios1 = nlios.copy(name='nlios1')
        nlios2 = nlios.copy(name='nlios2')

        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Single nonlinear system - no states
        ios_t, ios_y = ios.input_output_response(nlios, T, U)
        np.testing.assert_array_almost_equal(ios_y, U*U, decimal=3)

        # Composed nonlinear system (series)
        ios_t, ios_y = ios.input_output_response(nlios1 * nlios2, T, U)
        np.testing.assert_array_almost_equal(ios_y, U**4, decimal=3)

        # Composed nonlinear system (parallel)
        ios_t, ios_y = ios.input_output_response(nlios1 + nlios2, T, U)
        np.testing.assert_array_almost_equal(ios_y, 2*U**2, decimal=3)

        # Nonlinear system composed with LTI system (series) -- with states
        ios_t, ios_y = ios.input_output_response(
            nlios * lnios * nlios, T, U, X0)
        lti_t, lti_y = ct.forced_response(linsys, T, U*U, X0)
        np.testing.assert_array_almost_equal(ios_y, lti_y*lti_y, decimal=3)

        # Nonlinear system in feeback loop with LTI system
        iosys = ios.InterconnectedSystem(
            [lnios, nlios],         # linear system w/ nonlinear feedback
            [[1],                   # feedback interconnection (sig to 0)
             [0, (1, 0, -1)]],
            0,                      # input to linear system
            0                       # output from linear system
        )
        ios_t, ios_y = ios.input_output_response(iosys, T, U, X0)
        # No easy way to test the result

        # Algebraic loop from static nonlinear system in feedback
        # (error will be due to no states)
        iosys = ios.InterconnectedSystem(
            [nlios1, nlios2],       # two copies of a static nonlinear system
            [[0, 1],                # feedback interconnection
             [1, (0, 0, -1)]],
            0, 0
        )
        args = (iosys, T, U)
        with pytest.raises(RuntimeError):
            ios.input_output_response(*args)

        # Algebraic loop due to feedthrough term
        linsys = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0], [1]], [[1, 0]], [[1]])
        lnios = ios.LinearIOSystem(linsys)
        iosys = ios.InterconnectedSystem(
            [nlios, lnios],         # linear system w/ nonlinear feedback
            [[0, 1],                # feedback interconnection
             [1, (0, 0, -1)]],
            0, 0
        )
        args = (iosys, T, U, X0)
        # ios_t, ios_y = ios.input_output_response(iosys, T, U, X0)
        with pytest.raises(RuntimeError):
            ios.input_output_response(*args)

    def test_summer(self, tsys):
        # Construct a MIMO system for testing
        linsys = tsys.mimo_linsys1
        linio1 = ios.LinearIOSystem(linsys, name='linio1')
        linio2 = ios.LinearIOSystem(linsys, name='linio2')

        linsys_parallel = linsys + linsys
        iosys_parallel = linio1 + linio2

        # Set up parameters for simulation
        T = tsys.T
        U = [np.sin(T), np.cos(T)]
        X0 = 0

        lin_t, lin_y = ct.forced_response(linsys_parallel, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_parallel, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

    def test_rmul(self, tsys):
        # Test right multiplication
        # TODO: replace with better tests when conversions are implemented

        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Linear system with input and output nonlinearities
        # Also creates a nested interconnected system
        ioslin = ios.LinearIOSystem(tsys.siso_linsys)
        nlios =  ios.NonlinearIOSystem(None, \
            lambda t, x, u, params: u*u, inputs=1, outputs=1)
        sys1 = nlios * ioslin
        sys2 = ios.InputOutputSystem.__rmul__(nlios, sys1)

        # Make sure we got the right thing (via simulation comparison)
        ios_t, ios_y = ios.input_output_response(sys2, T, U, X0)
        lti_t, lti_y = ct.forced_response(ioslin, T, U*U, X0)
        np.testing.assert_array_almost_equal(ios_y, lti_y*lti_y, decimal=3)

    def test_neg(self, tsys):
        """Test negation of a system"""

        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Static nonlinear system
        nlios =  ios.NonlinearIOSystem(None, \
            lambda t, x, u, params: u*u, inputs=1, outputs=1)
        ios_t, ios_y = ios.input_output_response(-nlios, T, U)
        np.testing.assert_array_almost_equal(ios_y, -U*U, decimal=3)

        # Linear system with input nonlinearity
        # Also creates a nested interconnected system
        ioslin = ios.LinearIOSystem(tsys.siso_linsys)
        sys = (ioslin) * (-nlios)

        # Make sure we got the right thing (via simulation comparison)
        ios_t, ios_y = ios.input_output_response(sys, T, U, X0)
        lti_t, lti_y = ct.forced_response(ioslin, T, U*U, X0)
        np.testing.assert_array_almost_equal(ios_y, -lti_y, decimal=3)

    def test_feedback(self, tsys):
        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Linear system with constant feedback (via "nonlinear" mapping)
        ioslin = ios.LinearIOSystem(tsys.siso_linsys)
        nlios =  ios.NonlinearIOSystem(None, \
            lambda t, x, u, params: u, inputs=1, outputs=1)
        iosys = ct.feedback(ioslin, nlios)
        linsys = ct.feedback(tsys.siso_linsys, 1)

        ios_t, ios_y = ios.input_output_response(iosys, T, U, X0)
        lti_t, lti_y = ct.forced_response(linsys, T, U, X0)
        np.testing.assert_allclose(ios_y, lti_y,atol=0.002,rtol=0.)

    def test_bdalg_functions(self, tsys):
        """Test block diagram functions algebra on I/O systems"""
        # Set up parameters for simulation
        T = tsys.T
        U = [np.sin(T), np.cos(T)]
        X0 = 0

        # Set up systems to be composed
        linsys1 = tsys.mimo_linsys1
        linio1 = ios.LinearIOSystem(linsys1)
        linsys2 = tsys.mimo_linsys2
        linio2 = ios.LinearIOSystem(linsys2)

        # Series interconnection
        linsys_series = ct.series(linsys1, linsys2)
        iosys_series = ct.series(linio1, linio2)
        lin_t, lin_y = ct.forced_response(linsys_series, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_series, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Make sure that systems don't commute
        linsys_series = ct.series(linsys2, linsys1)
        lin_t, lin_y = ct.forced_response(linsys_series, T, U, X0)
        assert not (np.abs(lin_y - ios_y) < 1e-3).all()

        # Parallel interconnection
        linsys_parallel = ct.parallel(linsys1, linsys2)
        iosys_parallel = ct.parallel(linio1, linio2)
        lin_t, lin_y = ct.forced_response(linsys_parallel, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_parallel, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Negation
        linsys_negate = ct.negate(linsys1)
        iosys_negate = ct.negate(linio1)
        lin_t, lin_y = ct.forced_response(linsys_negate, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_negate, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Feedback interconnection
        linsys_feedback = ct.feedback(linsys1, linsys2)
        iosys_feedback = ct.feedback(linio1, linio2)
        lin_t, lin_y = ct.forced_response(linsys_feedback, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_feedback, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

    def test_algebraic_functions(self, tsys):
        """Test algebraic operations on I/O systems"""
        # Set up parameters for simulation
        T = tsys.T
        U = [np.sin(T), np.cos(T)]
        X0 = 0

        # Set up systems to be composed
        linsys1 = tsys.mimo_linsys1
        linio1 = ios.LinearIOSystem(linsys1)
        linsys2 = tsys.mimo_linsys2
        linio2 = ios.LinearIOSystem(linsys2)

        # Multiplication
        linsys_mul = linsys2 * linsys1
        iosys_mul = linio2 * linio1
        lin_t, lin_y = ct.forced_response(linsys_mul, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_mul, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Make sure that systems don't commute
        linsys_mul = linsys1 * linsys2
        lin_t, lin_y = ct.forced_response(linsys_mul, T, U, X0)
        assert not (np.abs(lin_y - ios_y) < 1e-3).all()

        # Addition
        linsys_add = linsys1 + linsys2
        iosys_add = linio1 + linio2
        lin_t, lin_y = ct.forced_response(linsys_add, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_add, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Subtraction
        linsys_sub = linsys1 - linsys2
        iosys_sub = linio1 - linio2
        lin_t, lin_y = ct.forced_response(linsys_sub, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_sub, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Make sure that systems don't commute
        linsys_sub = linsys2 - linsys1
        lin_t, lin_y = ct.forced_response(linsys_sub, T, U, X0)
        assert not (np.abs(lin_y - ios_y) < 1e-3).all()

        # Negation
        linsys_negate = -linsys1
        iosys_negate = -linio1
        lin_t, lin_y = ct.forced_response(linsys_negate, T, U, X0)
        ios_t, ios_y = ios.input_output_response(iosys_negate, T, U, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

    def test_nonsquare_bdalg(self, tsys):
        # Set up parameters for simulation
        T = tsys.T
        U2 = [np.sin(T), np.cos(T)]
        U3 = [np.sin(T), np.cos(T), T]
        X0 = 0

        # Set up systems to be composed
        linsys_2i3o = ct.StateSpace(
            [[-1, 1, 0], [0, -2, 0], [0, 0, -3]], [[1, 0], [0, 1], [1, 1]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], np.zeros((3, 2)))
        iosys_2i3o = ios.LinearIOSystem(linsys_2i3o)

        linsys_3i2o = ct.StateSpace(
            [[-1, 1, 0], [0, -2, 0], [0, 0, -3]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[1, 0, 1], [0, 1, -1]], np.zeros((2, 3)))
        iosys_3i2o = ios.LinearIOSystem(linsys_3i2o)

        # Multiplication
        linsys_multiply = linsys_3i2o * linsys_2i3o
        iosys_multiply = iosys_3i2o * iosys_2i3o
        lin_t, lin_y = ct.forced_response(linsys_multiply, T, U2, X0)
        ios_t, ios_y = ios.input_output_response(iosys_multiply, T, U2, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        linsys_multiply = linsys_2i3o * linsys_3i2o
        iosys_multiply = iosys_2i3o * iosys_3i2o
        lin_t, lin_y = ct.forced_response(linsys_multiply, T, U3, X0)
        ios_t, ios_y = ios.input_output_response(iosys_multiply, T, U3, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Right multiplication
        # TODO: add real tests once conversion from other types is supported
        iosys_multiply = ios.InputOutputSystem.__rmul__(iosys_3i2o, iosys_2i3o)
        ios_t, ios_y = ios.input_output_response(iosys_multiply, T, U3, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Feedback
        linsys_multiply = ct.feedback(linsys_3i2o, linsys_2i3o)
        iosys_multiply = iosys_3i2o.feedback(iosys_2i3o)
        lin_t, lin_y = ct.forced_response(linsys_multiply, T, U3, X0)
        ios_t, ios_y = ios.input_output_response(iosys_multiply, T, U3, X0)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Mismatch should generate exception
        args = (iosys_3i2o, iosys_3i2o)
        with pytest.raises(ValueError):
            ct.series(*args)

    def test_discrete(self, tsys):
        """Test discrete time functionality"""
        # Create some linear and nonlinear systems to play with
        linsys = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0], [1]], [[1, 0]], [[0]], True)
        lnios = ios.LinearIOSystem(linsys)

        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Simulate and compare to LTI output
        ios_t, ios_y = ios.input_output_response(lnios, T, U, X0)
        lin_t, lin_y = ct.forced_response(linsys, T, U, X0)
        np.testing.assert_allclose(ios_t, lin_t,atol=0.002,rtol=0.)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

        # Test MIMO system, converted to discrete time
        linsys = ct.StateSpace(tsys.mimo_linsys1)
        linsys.dt = tsys.T[1] - tsys.T[0]
        lnios = ios.LinearIOSystem(linsys)

        # Set up parameters for simulation
        T = tsys.T
        U = [np.sin(T), np.cos(T)]
        X0 = 0

        # Simulate and compare to LTI output
        ios_t, ios_y = ios.input_output_response(lnios, T, U, X0)
        lin_t, lin_y = ct.forced_response(linsys, T, U, X0)
        np.testing.assert_allclose(ios_t, lin_t,atol=0.002,rtol=0.)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

    def test_discrete_iosys(self, tsys):
        """Create a discrete time system from scratch"""
        linsys = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0], [1]], [[1, 0]], [[0]], True)

        # Create nonlinear version of the same system
        def nlsys_update(t, x, u, params):
            A, B = params['A'], params['B']
            return A @ x + B @ u
        def nlsys_output(t, x, u, params):
            C = params['C']
            return C @ x
        nlsys = ct.NonlinearIOSystem(
            nlsys_update, nlsys_output, inputs=1, outputs=1, states=2, dt=True)

        # Set up parameters for simulation
        T, U, X0 = tsys.T, tsys.U, tsys.X0

        # Simulate and compare to LTI output
        ios_t, ios_y = ios.input_output_response(
            nlsys, T, U, X0,
            params={'A': linsys.A, 'B': linsys.B, 'C': linsys.C})
        lin_t, lin_y = ct.forced_response(linsys, T, U, X0)
        np.testing.assert_allclose(ios_t, lin_t,atol=0.002,rtol=0.)
        np.testing.assert_allclose(ios_y, lin_y,atol=0.002,rtol=0.)

    def test_find_eqpts_dfan(self, tsys):
        """Test find_eqpt function on dfan example"""
        # Simple equilibrium point with no inputs
        nlsys = ios.NonlinearIOSystem(predprey)
        xeq, ueq, result = ios.find_eqpt(
            nlsys, [1.6, 1.2], None, return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(xeq, [1.64705879, 1.17923874])
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((2,)))

        # Ducted fan dynamics with output = velocity
        nlsys = ios.NonlinearIOSystem(pvtol, lambda t, x, u, params: x[0:2])

        # Make sure the origin is a fixed point
        xeq, ueq, result = ios.find_eqpt(
            nlsys, [0, 0, 0, 0], [0, 4*9.8], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((4,)))
        np.testing.assert_array_almost_equal(xeq, [0, 0, 0, 0])

        # Use a small lateral force to cause motion
        xeq, ueq, result = ios.find_eqpt(
            nlsys, [0, 0, 0, 0], [0.01, 4*9.8], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((4,)), decimal=5)

        # Equilibrium point with fixed output
        xeq, ueq, result = ios.find_eqpt(
            nlsys, [0, 0, 0, 0], [0.01, 4*9.8],
            y0=[0.1, 0.1], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys._out(0, xeq, ueq), [0.1, 0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((4,)), decimal=5)

        # Specify outputs to constrain (replicate previous)
        xeq, ueq, result = ios.find_eqpt(
            nlsys, [0, 0, 0, 0], [0.01, 4*9.8], y0=[0.1, 0.1],
            iy = [0, 1], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys._out(0, xeq, ueq), [0.1, 0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((4,)), decimal=5)

        # Specify inputs to constrain (replicate previous), w/ no result
        xeq, ueq = ios.find_eqpt(
            nlsys, [0, 0, 0, 0], [0.01, 4*9.8], y0=[0.1, 0.1], iu = [])
        np.testing.assert_array_almost_equal(
            nlsys._out(0, xeq, ueq), [0.1, 0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys._rhs(0, xeq, ueq), np.zeros((4,)), decimal=5)

        # Now solve the problem with the original PVTOL variables
        # Constrain the output angle and x velocity
        nlsys_full = ios.NonlinearIOSystem(pvtol_full, None)
        xeq, ueq, result = ios.find_eqpt(
            nlsys_full, [0, 0, 0, 0, 0, 0], [0.01, 4*9.8],
            y0=[0, 0, 0.1, 0.1, 0, 0], iy = [2, 3],
            idx=[2, 3, 4, 5], ix=[0, 1], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys_full._out(0, xeq, ueq)[[2, 3]], [0.1, 0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys_full._rhs(0, xeq, ueq)[-4:], np.zeros((4,)), decimal=5)

        # Same test as before, but now all constraints are in the state vector
        nlsys_full = ios.NonlinearIOSystem(pvtol_full, None)
        xeq, ueq, result = ios.find_eqpt(
            nlsys_full, [0, 0, 0.1, 0.1, 0, 0], [0.01, 4*9.8],
            idx=[2, 3, 4, 5], ix=[0, 1, 2, 3], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys_full._out(0, xeq, ueq)[[2, 3]], [0.1, 0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys_full._rhs(0, xeq, ueq)[-4:], np.zeros((4,)), decimal=5)

        # Fix one input and vary the other
        nlsys_full = ios.NonlinearIOSystem(pvtol_full, None)
        xeq, ueq, result = ios.find_eqpt(
            nlsys_full, [0, 0, 0, 0, 0, 0], [0.01, 4*9.8],
            y0=[0, 0, 0.1, 0.1, 0, 0], iy=[3], iu=[1],
            idx=[2, 3, 4, 5], ix=[0, 1], return_result=True)
        assert result.success
        np.testing.assert_almost_equal(ueq[1], 4*9.8, decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys_full._out(0, xeq, ueq)[[3]], [0.1], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys_full._rhs(0, xeq, ueq)[-4:], np.zeros((4,)), decimal=5)

        # PVTOL with output = y velocity
        xeq, ueq, result = ios.find_eqpt(
            nlsys_full, [0, 0, 0, 0.1, 0, 0], [0.01, 4*9.8],
            y0=[0, 0, 0, 0.1, 0, 0], iy=[3],
            dx0=[0.1, 0, 0, 0, 0, 0], idx=[1, 2, 3, 4, 5],
            ix=[0, 1], return_result=True)
        assert result.success
        np.testing.assert_array_almost_equal(
            nlsys_full._out(0, xeq, ueq)[-3:], [0.1, 0, 0], decimal=5)
        np.testing.assert_array_almost_equal(
            nlsys_full._rhs(0, xeq, ueq)[-5:], np.zeros((5,)), decimal=5)

        # Unobservable system
        linsys = ct.StateSpace(
            [[-1, 1], [0, -2]], [[0], [1]], [[0, 0]], [[0]])
        lnios = ios.LinearIOSystem(linsys)

        # If result is returned, user has to check
        xeq, ueq, result = ios.find_eqpt(
            lnios, [0, 0], [0], y0=[1], return_result=True)
        assert not result.success

        # If result is not returned, find_eqpt should return None
        xeq, ueq = ios.find_eqpt(lnios, [0, 0], [0], y0=[1])
        assert xeq is None
        assert ueq is None

    def test_params(self, tsys):
        # Start with the default set of parameters
        ios_secord_default = ios.NonlinearIOSystem(
            secord_update, secord_output, inputs=1, outputs=1, states=2)
        lin_secord_default = ios.linearize(ios_secord_default, [0, 0], [0])
        w_default, v_default = np.linalg.eig(lin_secord_default.A)

        # New copy, with modified parameters
        ios_secord_update = ios.NonlinearIOSystem(
            secord_update, secord_output, inputs=1, outputs=1, states=2,
            params={'omega0':2, 'zeta':0})

        # Make sure the default parameters haven't changed
        lin_secord_check = ios.linearize(ios_secord_default, [0, 0], [0])
        w, v = np.linalg.eig(lin_secord_check.A)
        np.testing.assert_array_almost_equal(np.sort(w), np.sort(w_default))

        # Make sure updated system parameters got set correctly
        lin_secord_update = ios.linearize(ios_secord_update, [0, 0], [0])
        w, v = np.linalg.eig(lin_secord_update.A)
        np.testing.assert_array_almost_equal(np.sort(w), np.sort([2j, -2j]))

        # Change the parameters of the default sys just for the linearization
        lin_secord_local = ios.linearize(ios_secord_default, [0, 0], [0],
                                          params={'zeta':0})
        w, v = np.linalg.eig(lin_secord_local.A)
        np.testing.assert_array_almost_equal(np.sort(w), np.sort([1j, -1j]))

        # Change the parameters of the updated sys just for the linearization
        lin_secord_local = ios.linearize(ios_secord_update, [0, 0], [0],
                                          params={'zeta':0, 'omega0':3})
        w, v = np.linalg.eig(lin_secord_local.A)
        np.testing.assert_array_almost_equal(np.sort(w), np.sort([3j, -3j]))

        # Make sure that changes propagate through interconnections
        ios_series_default_local = ios_secord_default * ios_secord_update
        lin_series_default_local = ios.linearize(
            ios_series_default_local, [0, 0, 0, 0], [0])
        w, v = np.linalg.eig(lin_series_default_local.A)
        np.testing.assert_array_almost_equal(
            np.sort(w), np.sort(np.concatenate((w_default, [2j, -2j]))))

        # Show that we can change the parameters at linearization
        lin_series_override = ios.linearize(
            ios_series_default_local, [0, 0, 0, 0], [0],
            params={'zeta':0, 'omega0':4})
        w, v = np.linalg.eig(lin_series_override.A)
        np.testing.assert_array_almost_equal(w, [4j, -4j, 4j, -4j])

        # Check for warning if we try to set params for LinearIOSystem
        linsys = tsys.siso_linsys
        iosys = ios.LinearIOSystem(linsys)
        T, U, X0 = tsys.T, tsys.U, tsys.X0
        lti_t, lti_y = ct.forced_response(linsys, T, U, X0)
        with pytest.warns(UserWarning, match="LinearIOSystem.*ignored"):
            ios_t, ios_y = ios.input_output_response(
                iosys, T, U, X0, params={'something':0})

        # Check to make sure results are OK
        np.testing.assert_array_almost_equal(lti_t, ios_t)
        np.testing.assert_allclose(lti_y, ios_y,atol=0.002,rtol=0.)

    def test_named_signals(self, tsys):
        sys1 = ios.NonlinearIOSystem(
            updfcn = lambda t, x, u, params: np.array(
                tsys.mimo_linsys1.A @ np.reshape(x, (-1, 1)) \
                + tsys.mimo_linsys1.B @ np.reshape(u, (-1, 1))
            ).reshape(-1,),
            outfcn = lambda t, x, u, params: np.array(
                tsys.mimo_linsys1.C @ np.reshape(x, (-1, 1)) \
                + tsys.mimo_linsys1.D @ np.reshape(u, (-1, 1))
            ).reshape(-1,),
            inputs = ['u[0]', 'u[1]'],
            outputs = ['y[0]', 'y[1]'],
            states = tsys.mimo_linsys1.nstates,
            name = 'sys1')
        sys2 = ios.LinearIOSystem(tsys.mimo_linsys2,
            inputs = ['u[0]', 'u[1]'],
            outputs = ['y[0]', 'y[1]'],
            name = 'sys2')

        # Series interconnection (sys1 * sys2) using __mul__
        ios_mul = sys1 * sys2
        ss_series = tsys.mimo_linsys1 * tsys.mimo_linsys2
        lin_series = ct.linearize(ios_mul, 0, 0)
        np.testing.assert_array_almost_equal(ss_series.A, lin_series.A)
        np.testing.assert_array_almost_equal(ss_series.B, lin_series.B)
        np.testing.assert_array_almost_equal(ss_series.C, lin_series.C)
        np.testing.assert_array_almost_equal(ss_series.D, lin_series.D)

        # Series interconnection (sys1 * sys2) using series
        ios_series = ct.series(sys2, sys1)
        ss_series = ct.series(tsys.mimo_linsys2, tsys.mimo_linsys1)
        lin_series = ct.linearize(ios_series, 0, 0)
        np.testing.assert_array_almost_equal(ss_series.A, lin_series.A)
        np.testing.assert_array_almost_equal(ss_series.B, lin_series.B)
        np.testing.assert_array_almost_equal(ss_series.C, lin_series.C)
        np.testing.assert_array_almost_equal(ss_series.D, lin_series.D)

        # Series interconnection (sys1 * sys2) using named + mixed signals
        ios_connect = ios.InterconnectedSystem(
            [sys2, sys1],
            connections=[
                [('sys1', 'u[0]'), 'sys2.y[0]'],
                ['sys1.u[1]', 'sys2.y[1]']
            ],
            inplist=['sys2.u[0]', ('sys2', 1)],
            outlist=[(1, 'y[0]'), 'sys1.y[1]']
        )
        lin_series = ct.linearize(ios_connect, 0, 0)
        np.testing.assert_array_almost_equal(ss_series.A, lin_series.A)
        np.testing.assert_array_almost_equal(ss_series.B, lin_series.B)
        np.testing.assert_array_almost_equal(ss_series.C, lin_series.C)
        np.testing.assert_array_almost_equal(ss_series.D, lin_series.D)

        # Try the same thing using the interconnect function
        # Since sys1 is nonlinear, we should get back the same result
        ios_connect = ios.interconnect(
            (sys2, sys1),
            connections=(
                [('sys1', 'u[0]'), 'sys2.y[0]'],
                ['sys1.u[1]', 'sys2.y[1]']
            ),
            inplist=['sys2.u[0]', ('sys2', 1)],
            outlist=[(1, 'y[0]'), 'sys1.y[1]']
        )
        lin_series = ct.linearize(ios_connect, 0, 0)
        np.testing.assert_array_almost_equal(ss_series.A, lin_series.A)
        np.testing.assert_array_almost_equal(ss_series.B, lin_series.B)
        np.testing.assert_array_almost_equal(ss_series.C, lin_series.C)
        np.testing.assert_array_almost_equal(ss_series.D, lin_series.D)

        # Try the same thing using the interconnect function
        # Since sys1 is nonlinear, we should get back the same result
        # Note: use a tuple for connections to make sure it works
        ios_connect = ios.interconnect(
            (sys2, sys1),
            connections=(
                [('sys1', 'u[0]'), 'sys2.y[0]'],
                ['sys1.u[1]', 'sys2.y[1]']
            ),
            inplist=['sys2.u[0]', ('sys2', 1)],
            outlist=[(1, 'y[0]'), 'sys1.y[1]']
        )
        lin_series = ct.linearize(ios_connect, 0, 0)
        np.testing.assert_array_almost_equal(ss_series.A, lin_series.A)
        np.testing.assert_array_almost_equal(ss_series.B, lin_series.B)
        np.testing.assert_array_almost_equal(ss_series.C, lin_series.C)
        np.testing.assert_array_almost_equal(ss_series.D, lin_series.D)

        # Make sure that we can use input signal names as system outputs
        ios_connect = ios.InterconnectedSystem(
            [sys1, sys2],
            connections=[
                ['sys2.u[0]', 'sys1.y[0]'], ['sys2.u[1]', 'sys1.y[1]'],
                ['sys1.u[0]', '-sys2.y[0]'], ['sys1.u[1]', '-sys2.y[1]']
            ],
            inplist=['sys1.u[0]', 'sys1.u[1]'],
            outlist=['sys2.u[0]', 'sys2.u[1]']  # = sys1.y[0], sys1.y[1]
        )
        ss_feedback = ct.feedback(tsys.mimo_linsys1, tsys.mimo_linsys2)
        lin_feedback = ct.linearize(ios_connect, 0, 0)
        np.testing.assert_array_almost_equal(ss_feedback.A, lin_feedback.A)
        np.testing.assert_array_almost_equal(ss_feedback.B, lin_feedback.B)
        np.testing.assert_array_almost_equal(ss_feedback.C, lin_feedback.C)
        np.testing.assert_array_almost_equal(ss_feedback.D, lin_feedback.D)

    @pytest.mark.usefixtures("editsdefaults")
    def test_sys_naming_convention(self, tsys):
        """Enforce generic system names 'sys[i]' to be present when systems are
        created without explicit names."""

        ct.config.use_legacy_defaults('0.8.4')  # changed delims in 0.9.0
        ct.config.use_numpy_matrix(False)       # np.matrix deprecated

        # Create a system with a known ID
        ct.namedio.NamedIOSystem._idCounter = 0
        sys = ct.ss(
            tsys.mimo_linsys1.A, tsys.mimo_linsys1.B,
            tsys.mimo_linsys1.C, tsys.mimo_linsys1.D)

        assert sys.name == "sys[0]"
        assert sys.copy().name == "copy of sys[0]"

        namedsys = ios.NonlinearIOSystem(
            updfcn=lambda t, x, u, params: x,
            outfcn=lambda t, x, u, params: u,
            inputs=('u[0]', 'u[1]'),
            outputs=('y[0]', 'y[1]'),
            states=tsys.mimo_linsys1.nstates,
            name='namedsys')
        unnamedsys1 = ct.NonlinearIOSystem(
            lambda t, x, u, params: x, inputs=2, outputs=2, states=2
        )
        unnamedsys2 = ct.NonlinearIOSystem(
            None, lambda t, x, u, params: u, inputs=2, outputs=2
        )
        assert unnamedsys2.name == "sys[2]"

        # Unnamed/unnamed connections
        uu_series = unnamedsys1 * unnamedsys2
        uu_parallel = unnamedsys1 + unnamedsys2
        u_neg = - unnamedsys1
        uu_feedback = unnamedsys2.feedback(unnamedsys1)
        uu_dup = unnamedsys1 * unnamedsys1.copy()
        uu_hierarchical = uu_series * unnamedsys1

        assert uu_series.name == "sys[3]"
        assert uu_parallel.name == "sys[4]"
        assert u_neg.name == "sys[5]"
        assert uu_feedback.name == "sys[6]"
        assert uu_dup.name == "sys[7]"
        assert uu_hierarchical.name == "sys[8]"

        # Unnamed/named connections
        un_series = unnamedsys1 * namedsys
        un_parallel = unnamedsys1 + namedsys
        un_feedback = unnamedsys2.feedback(namedsys)
        un_dup = unnamedsys1 * namedsys.copy()
        un_hierarchical = uu_series * unnamedsys1

        assert un_series.name == "sys[9]"
        assert un_parallel.name == "sys[10]"
        assert un_feedback.name == "sys[11]"
        assert un_dup.name == "sys[12]"
        assert un_hierarchical.name == "sys[13]"

        # Same system conflict
        with pytest.warns(UserWarning):
            namedsys * namedsys

    @pytest.mark.usefixtures("editsdefaults")
    def test_signals_naming_convention_0_8_4(self, tsys):
        """Enforce generic names to be present when systems are created
        without explicit signal names:
        input: 'u[i]'
        state: 'x[i]'
        output: 'y[i]'
        """

        ct.config.use_legacy_defaults('0.8.4')  # changed delims in 0.9.0
        ct.config.use_numpy_matrix(False)       # np.matrix deprecated

        # Create a system with a known ID
        ct.namedio.NamedIOSystem._idCounter = 0
        sys = ct.ss(
            tsys.mimo_linsys1.A, tsys.mimo_linsys1.B,
            tsys.mimo_linsys1.C, tsys.mimo_linsys1.D)

        for statename in ["x[0]", "x[1]"]:
            assert statename in sys.state_index
        for inputname in ["u[0]", "u[1]"]:
            assert inputname in sys.input_index
        for outputname in ["y[0]", "y[1]"]:
            assert outputname in sys.output_index
        assert len(sys.state_index) == sys.nstates
        assert len(sys.input_index) == sys.ninputs
        assert len(sys.output_index) == sys.noutputs

        namedsys = ios.NonlinearIOSystem(
            updfcn=lambda t, x, u, params: x,
            outfcn=lambda t, x, u, params: u,
            inputs=('u0'),
            outputs=('y0'),
            states=('x0'),
            name='namedsys')
        unnamedsys = ct.NonlinearIOSystem(
            lambda t, x, u, params: x, inputs=1, outputs=1, states=1
        )
        assert 'u0' in namedsys.input_index
        assert 'y0' in namedsys.output_index
        assert 'x0' in namedsys.state_index

        # Unnamed/named connections
        un_series = unnamedsys * namedsys
        un_parallel = unnamedsys + namedsys
        un_feedback = unnamedsys.feedback(namedsys)
        un_dup = unnamedsys * namedsys.copy()
        un_hierarchical = un_series*unnamedsys
        u_neg = - unnamedsys

        assert "sys[1].x[0]" in un_series.state_index
        assert "namedsys.x0" in un_series.state_index
        assert "sys[1].x[0]" in un_parallel.state_index
        assert "namedsys.x0" in un_series.state_index
        assert "sys[1].x[0]" in un_feedback.state_index
        assert "namedsys.x0" in un_feedback.state_index
        assert "sys[1].x[0]" in un_dup.state_index
        assert "copy of namedsys.x0" in un_dup.state_index
        assert "sys[1].x[0]" in un_hierarchical.state_index
        assert "sys[2].sys[1].x[0]" in un_hierarchical.state_index
        assert "sys[1].x[0]" in u_neg.state_index

        # Same system conflict
        with pytest.warns(UserWarning):
            same_name_series = namedsys * namedsys
            assert "namedsys.x0" in same_name_series.state_index
            assert "copy of namedsys.x0" in same_name_series.state_index

    def test_named_signals_linearize_inconsistent(self, tsys):
        """Make sure that providing inputs or outputs not consistent with
           updfcn or outfcn fail
        """

        def updfcn(t, x, u, params):
            """2 inputs, 2 states"""
            return np.array(
                tsys.mimo_linsys1.A @ np.reshape(x, (-1, 1))
                + tsys.mimo_linsys1.B @ np.reshape(u, (-1, 1))
                ).reshape(-1,)

        def outfcn(t, x, u, params):
            """2 states, 2 outputs"""
            return np.array(
                    tsys.mimo_linsys1.C * np.reshape(x, (-1, 1))
                    + tsys.mimo_linsys1.D * np.reshape(u, (-1, 1))
                ).reshape(-1,)

        for inputs, outputs in [
                (('u[0]'), ('y[0]', 'y[1]')),  # not enough u
                (('u[0]', 'u[1]', 'u[toomuch]'), ('y[0]', 'y[1]')),
                (('u[0]', 'u[1]'), ('y[0]')),  # not enough y
                (('u[0]', 'u[1]'), ('y[0]', 'y[1]', 'y[toomuch]'))]:
            sys1 = ios.NonlinearIOSystem(updfcn=updfcn,
                                         outfcn=outfcn,
                                         inputs=inputs,
                                         outputs=outputs,
                                         states=tsys.mimo_linsys1.nstates,
                                         name='sys1')
            with pytest.raises(ValueError):
                sys1.linearize([0, 0], [0, 0])

        sys2 = ios.NonlinearIOSystem(updfcn=updfcn,
                                     outfcn=outfcn,
                                     inputs=('u[0]', 'u[1]'),
                                     outputs=('y[0]', 'y[1]'),
                                     states=tsys.mimo_linsys1.nstates,
                                     name='sys1')
        for x0, u0 in [([0], [0, 0]),
                       ([0, 0, 0], [0, 0]),
                       ([0, 0], [0]),
                       ([0, 0], [0, 0, 0])]:
            with pytest.raises(ValueError):
                sys2.linearize(x0, u0)

    def test_linearize_concatenation(self, kincar):
        # Create a simple nonlinear system to check (kinematic car)
        iosys = kincar
        linearized = iosys.linearize([0, np.array([0, 0])], [0, 0])
        np.testing.assert_array_almost_equal(linearized.A, np.zeros((3,3)))
        np.testing.assert_array_almost_equal(
            linearized.B, [[1, 0], [0, 0], [0, 1]])
        np.testing.assert_array_almost_equal(
            linearized.C, [[1, 0, 0], [0, 1, 0]])
        np.testing.assert_array_almost_equal(linearized.D, np.zeros((2,2)))

    def test_lineariosys_statespace(self, tsys):
        """Make sure that a LinearIOSystem is also a StateSpace object"""
        iosys_siso = ct.LinearIOSystem(tsys.siso_linsys, name='siso')
        iosys_siso2 = ct.LinearIOSystem(tsys.siso_linsys, name='siso2')
        assert isinstance(iosys_siso, ct.StateSpace)

        # Make sure that state space functions work for LinearIOSystems
        np.testing.assert_allclose(
            iosys_siso.poles(), tsys.siso_linsys.poles())
        omega = np.logspace(.1, 10, 100)
        mag_io, phase_io, omega_io = iosys_siso.frequency_response(omega)
        mag_ss, phase_ss, omega_ss = tsys.siso_linsys.frequency_response(omega)
        np.testing.assert_allclose(mag_io, mag_ss)
        np.testing.assert_allclose(phase_io, phase_ss)
        np.testing.assert_allclose(omega_io, omega_ss)

        # LinearIOSystem methods should override StateSpace methods
        io_mul = iosys_siso * iosys_siso2
        assert isinstance(io_mul, ct.InputOutputSystem)

        # But also retain linear structure
        assert isinstance(io_mul, ct.StateSpace)

        # And make sure the systems match
        ss_series = tsys.siso_linsys * tsys.siso_linsys
        np.testing.assert_allclose(io_mul.A, ss_series.A)
        np.testing.assert_allclose(io_mul.B, ss_series.B)
        np.testing.assert_allclose(io_mul.C, ss_series.C)
        np.testing.assert_allclose(io_mul.D, ss_series.D)

        # Make sure that series does the same thing
        io_series = ct.series(iosys_siso, iosys_siso2)
        assert isinstance(io_series, ct.InputOutputSystem)
        assert isinstance(io_series, ct.StateSpace)
        np.testing.assert_allclose(io_series.A, ss_series.A)
        np.testing.assert_allclose(io_series.B, ss_series.B)
        np.testing.assert_allclose(io_series.C, ss_series.C)
        np.testing.assert_allclose(io_series.D, ss_series.D)

        # Test out feedback as well
        io_feedback = ct.feedback(iosys_siso, iosys_siso2)
        assert isinstance(io_series, ct.InputOutputSystem)

        # But also retain linear structure
        assert isinstance(io_series, ct.StateSpace)

        # And make sure the systems match
        ss_feedback = ct.feedback(tsys.siso_linsys, tsys.siso_linsys)
        np.testing.assert_allclose(io_feedback.A, ss_feedback.A)
        np.testing.assert_allclose(io_feedback.B, ss_feedback.B)
        np.testing.assert_allclose(io_feedback.C, ss_feedback.C)
        np.testing.assert_allclose(io_feedback.D, ss_feedback.D)

        # Make sure series interconnections are done in the right order
        ss_sys1 = ct.rss(2, 3, 2)
        io_sys1 = ct.ss2io(ss_sys1)
        ss_sys2 = ct.rss(2, 2, 3)
        io_sys2 = ct.ss2io(ss_sys2)
        io_series = io_sys2 * io_sys1
        assert io_series.ninputs == 2
        assert io_series.noutputs == 2
        assert io_series.nstates == 4

        # While we are at it, check that the state space matrices match
        ss_series = ss_sys2 * ss_sys1
        np.testing.assert_allclose(io_series.A, ss_series.A)
        np.testing.assert_allclose(io_series.B, ss_series.B)
        np.testing.assert_allclose(io_series.C, ss_series.C)
        np.testing.assert_allclose(io_series.D, ss_series.D)

    @pytest.mark.parametrize(
        "Pout, Pin, C, op, PCout, PCin", [
            (2, 2, 'rss', ct.LinearIOSystem.__mul__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__mul__, 2, 2),
            (2, 3, 2, ct.LinearIOSystem.__mul__, 2, 3),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__mul__, 2, 2),
            (2, 2, 'rss', ct.LinearIOSystem.__rmul__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__rmul__, 2, 2),
            (2, 3, 2, ct.LinearIOSystem.__rmul__, 2, 3),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__rmul__, 2, 2),
            (2, 2, 'rss', ct.LinearIOSystem.__add__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__add__, 2, 2),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__add__, 2, 2),
            (2, 2, 'rss', ct.LinearIOSystem.__radd__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__radd__, 2, 2),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__radd__, 2, 2),
            (2, 2, 'rss', ct.LinearIOSystem.__sub__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__sub__, 2, 2),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__sub__, 2, 2),
            (2, 2, 'rss', ct.LinearIOSystem.__rsub__, 2, 2),
            (2, 2, 2, ct.LinearIOSystem.__rsub__, 2, 2),
            (2, 2, np.random.rand(2, 2), ct.LinearIOSystem.__rsub__, 2, 2),

        ])
    def test_operand_conversion(self, Pout, Pin, C, op, PCout, PCin):
        P = ct.LinearIOSystem(
            ct.rss(2, Pout, Pin, strictly_proper=True), name='P')
        if isinstance(C, str) and C == 'rss':
            # Need to generate inside class to avoid matrix deprecation error
            C = ct.rss(2, 2, 2)
        PC = op(P, C)
        assert isinstance(PC, ct.LinearIOSystem)
        assert isinstance(PC, ct.StateSpace)
        assert PC.noutputs == PCout
        assert PC.ninputs == PCin

    @pytest.mark.parametrize(
        "Pout, Pin, C, op", [
            (2, 2, 'rss32', ct.LinearIOSystem.__mul__),
            (2, 2, 'rss23', ct.LinearIOSystem.__rmul__),
            (2, 2, 'rss32', ct.LinearIOSystem.__add__),
            (2, 2, 'rss23', ct.LinearIOSystem.__radd__),
            (2, 3, 2, ct.LinearIOSystem.__add__),
            (2, 3, 2, ct.LinearIOSystem.__radd__),
            (2, 2, 'rss32', ct.LinearIOSystem.__sub__),
            (2, 2, 'rss23', ct.LinearIOSystem.__rsub__),
            (2, 3, 2, ct.LinearIOSystem.__sub__),
            (2, 3, 2, ct.LinearIOSystem.__rsub__),
        ])
    def test_operand_incompatible(self, Pout, Pin, C, op):
        P = ct.LinearIOSystem(
            ct.rss(2, Pout, Pin, strictly_proper=True), name='P')
        if isinstance(C, str) and C == 'rss32':
            C = ct.rss(2, 3, 2)
        elif isinstance(C, str) and C == 'rss23':
            C = ct.rss(2, 2, 3)
        with pytest.raises(ValueError, match="incompatible"):
            PC = op(P, C)

    @pytest.mark.parametrize(
        "C, op", [
            (None, ct.LinearIOSystem.__mul__),
            (None, ct.LinearIOSystem.__rmul__),
            (None, ct.LinearIOSystem.__add__),
            (None, ct.LinearIOSystem.__radd__),
            (None, ct.LinearIOSystem.__sub__),
            (None, ct.LinearIOSystem.__rsub__),
        ])
    def test_operand_badtype(self, C, op):
        P = ct.LinearIOSystem(
            ct.rss(2, 2, 2, strictly_proper=True), name='P')
        with pytest.raises(TypeError, match="Unknown"):
            op(P, C)

    def test_neg_badsize(self):
        # Create a system of unspecified size
        sys = ct.InputOutputSystem()
        with pytest.raises(ValueError, match="Can't determine"):
            -sys

    def test_bad_signal_list(self):
        # Create a ystem with a bad signal list
        with pytest.raises(TypeError, match="Can't parse"):
            ct.InputOutputSystem(inputs=[1, 2, 3])

    def test_docstring_example(self):
        P = ct.LinearIOSystem(
            ct.rss(2, 2, 2, strictly_proper=True), name='P')
        C = ct.LinearIOSystem(ct.rss(2, 2, 2), name='C')
        S = ct.InterconnectedSystem(
            [C, P],
            connections = [
              ['P.u[0]', 'C.y[0]'], ['P.u[1]', 'C.y[1]'],
              ['C.u[0]', '-P.y[0]'], ['C.u[1]', '-P.y[1]']],
            inplist = ['C.u[0]', 'C.u[1]'],
            outlist = ['P.y[0]', 'P.y[1]'],
        )
        ss_P = ct.StateSpace(P.linearize(0, 0))
        ss_C = ct.StateSpace(C.linearize(0, 0))
        ss_eye = ct.StateSpace(
            [], np.zeros((0, 2)), np.zeros((2, 0)), np.eye(2))
        ss_S = ct.feedback(ss_P * ss_C, ss_eye)
        io_S = S.linearize(0, 0)
        np.testing.assert_array_almost_equal(io_S.A, ss_S.A)
        np.testing.assert_array_almost_equal(io_S.B, ss_S.B)
        np.testing.assert_array_almost_equal(io_S.C, ss_S.C)
        np.testing.assert_array_almost_equal(io_S.D, ss_S.D)

    @pytest.mark.usefixtures("editsdefaults")
    def test_duplicates(self, tsys):
        nlios = ios.NonlinearIOSystem(lambda t, x, u, params: x,
                                      lambda t, x, u, params: u * u,
                                      inputs=1, outputs=1, states=1,
                                      name="sys")

        # Duplicate objects
        with pytest.warns(UserWarning, match="duplicate object"):
            ios_series = nlios * nlios

        # Nonduplicate objects
        ct.config.use_legacy_defaults('0.8.4')  # changed delims in 0.9.0
        ct.config.use_numpy_matrix(False)       # np.matrix deprecated
        nlios1 = nlios.copy()
        nlios2 = nlios.copy()
        with pytest.warns(UserWarning, match="duplicate name"):
            ios_series = nlios1 * nlios2
            assert "copy of sys_1.x[0]" in ios_series.state_index.keys()
            assert "copy of sys.x[0]" in ios_series.state_index.keys()

        # Duplicate names
        iosys_siso = ct.LinearIOSystem(tsys.siso_linsys)
        nlios1 = ios.NonlinearIOSystem(None,
                                       lambda t, x, u, params: u * u,
                                       inputs=1, outputs=1, name="sys")
        nlios2 = ios.NonlinearIOSystem(None,
                                       lambda t, x, u, params: u * u,
                                       inputs=1, outputs=1, name="sys")

        with pytest.warns(UserWarning, match="duplicate name"):
            ct.InterconnectedSystem([nlios1, iosys_siso, nlios2],
                                    inputs=0, outputs=0, states=0)

        # Same system, different names => everything should be OK
        nlios1 = ios.NonlinearIOSystem(None,
                                       lambda t, x, u, params:  u * u,
                                       inputs=1, outputs=1, name="nlios1")
        nlios2 = ios.NonlinearIOSystem(None,
                                       lambda t, x, u, params: u * u,
                                       inputs=1, outputs=1, name="nlios2")
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            ct.InterconnectedSystem([nlios1, iosys_siso, nlios2],
                                    inputs=0, outputs=0, states=0)


def test_linear_interconnection():
    ss_sys1 = ct.rss(2, 2, 2, strictly_proper=True)
    ss_sys2 = ct.rss(2, 2, 2)
    io_sys1 = ios.LinearIOSystem(
        ss_sys1, inputs = ('u[0]', 'u[1]'),
        outputs = ('y[0]', 'y[1]'), name = 'sys1')
    io_sys2 = ios.LinearIOSystem(
        ss_sys2, inputs = ('u[0]', 'u[1]'),
        outputs = ('y[0]', 'y[1]'), name = 'sys2')
    nl_sys2 = ios.NonlinearIOSystem(
        lambda t, x, u, params: np.array(
            ss_sys2.A @ np.reshape(x, (-1, 1)) \
            + ss_sys2.B @ np.reshape(u, (-1, 1))
            ).reshape((-1,)),
        lambda t, x, u, params: np.array(
            ss_sys2.C @ np.reshape(x, (-1, 1)) \
            + ss_sys2.D @ np.reshape(u, (-1, 1))
            ).reshape((-1,)),
        states = 2,
        inputs = ('u[0]', 'u[1]'),
        outputs = ('y[0]', 'y[1]'),
        name = 'sys2')
    tf_siso = ct.tf(1, [0.1, 1])
    ss_siso = ct.ss(1, 2, 1, 1)
    nl_siso = ios.NonlinearIOSystem(
        lambda t, x, u, params: x*x,
        lambda t, x, u, params: u*x, states=1, inputs=1, outputs=1)

    # Create a "regular" InterconnectedSystem
    nl_connect = ios.interconnect(
        (io_sys1, nl_sys2),
        connections=[
            ['sys1.u[1]', 'sys2.y[0]'],
            ['sys2.u[0]', 'sys1.y[1]']
        ],
        inplist=[
            ['sys1.u[0]', 'sys1.u[1]'],
            ['sys2.u[1]']],
        outlist=[
            ['sys1.y[0]', '-sys2.y[0]'],
            ['sys2.y[1]'],
            ['sys2.u[1]']])
    assert isinstance(nl_connect, ios.InterconnectedSystem)
    assert not isinstance(nl_connect, ios.LinearICSystem)

    # Now take its linearization
    ss_connect = nl_connect.linearize(0, 0)
    assert isinstance(ss_connect, ios.LinearIOSystem)

    io_connect = ios.interconnect(
        (io_sys1, io_sys2),
        connections=[
            ['sys1.u[1]', 'sys2.y[0]'],
            ['sys2.u[0]', 'sys1.y[1]']
        ],
        inplist=[
            ['sys1.u[0]', 'sys1.u[1]'],
            ['sys2.u[1]']],
        outlist=[
            ['sys1.y[0]', '-sys2.y[0]'],
            ['sys2.y[1]'],
            ['sys2.u[1]']])
    assert isinstance(io_connect, ios.InterconnectedSystem)
    assert isinstance(io_connect, ios.LinearICSystem)
    assert isinstance(io_connect, ios.LinearIOSystem)
    assert isinstance(io_connect, ct.StateSpace)

    # Finally compare the linearization with the linear system
    np.testing.assert_array_almost_equal(io_connect.A, ss_connect.A)
    np.testing.assert_array_almost_equal(io_connect.B, ss_connect.B)
    np.testing.assert_array_almost_equal(io_connect.C, ss_connect.C)
    np.testing.assert_array_almost_equal(io_connect.D, ss_connect.D)

    # make sure interconnections of linear systems are linear and
    # if a nonlinear system is included then system is nonlinear
    assert isinstance(ss_siso*ss_siso, ios.LinearIOSystem)
    assert isinstance(tf_siso*ss_siso, ios.LinearIOSystem)
    assert isinstance(ss_siso*tf_siso, ios.LinearIOSystem)
    assert ~isinstance(ss_siso*nl_siso, ios.LinearIOSystem)
    assert ~isinstance(nl_siso*ss_siso, ios.LinearIOSystem)
    assert ~isinstance(nl_siso*nl_siso, ios.LinearIOSystem)
    assert ~isinstance(tf_siso*nl_siso, ios.LinearIOSystem)
    assert ~isinstance(nl_siso*tf_siso, ios.LinearIOSystem)
    assert ~isinstance(nl_siso*nl_siso, ios.LinearIOSystem)


def predprey(t, x, u, params={}):
    """Predator prey dynamics"""
    r = params.get('r', 2)
    d = params.get('d', 0.7)
    b = params.get('b', 0.3)
    k = params.get('k', 10)
    a = params.get('a', 8)
    c = params.get('c', 4)

    # Dynamics for the system
    dx0 = r * x[0] * (1 - x[0]/k) - a * x[1] * x[0]/(c + x[0])
    dx1 = b * a * x[1] * x[0] / (c + x[0]) - d * x[1]

    return np.array([dx0, dx1])


def pvtol(t, x, u, params={}):
    """Reduced planar vertical takeoff and landing dynamics"""
    from math import cos, sin
    m = params.get('m', 4.)      # kg, system mass
    J = params.get('J', 0.0475)  # kg m^2, system inertia
    r = params.get('r', 0.25)    # m, thrust offset
    g = params.get('g', 9.8)     # m/s, gravitational constant
    c = params.get('c', 0.05)    # N s/m, rotational damping
    l = params.get('c', 0.1)     # m, pivot location
    return np.array([
        x[3],
        -c/m * x[1] + 1/m * cos(x[0]) * u[0] - 1/m * sin(x[0]) * u[1],
        -g - c/m * x[2] + 1/m * sin(x[0]) * u[0] + 1/m * cos(x[0]) * u[1],
        -l/J * sin(x[0]) + r/J * u[0]
    ])


def pvtol_full(t, x, u, params={}):
    from math import cos, sin
    m = params.get('m', 4.)      # kg, system mass
    J = params.get('J', 0.0475)  # kg m^2, system inertia
    r = params.get('r', 0.25)    # m, thrust offset
    g = params.get('g', 9.8)     # m/s, gravitational constant
    c = params.get('c', 0.05)    # N s/m, rotational damping
    l = params.get('c', 0.1)     # m, pivot location
    return np.array([
        x[3], x[4], x[5],
        -c/m * x[3] + 1/m * cos(x[2]) * u[0] - 1/m * sin(x[2]) * u[1],
        -g - c/m * x[4] + 1/m * sin(x[2]) * u[0] + 1/m * cos(x[2]) * u[1],
        -l/J * sin(x[2]) + r/J * u[0]
    ])


def secord_update(t, x, u, params={}):
    """Second order system dynamics"""
    omega0 = params.get('omega0', 1.)
    zeta = params.get('zeta', 0.5)
    return np.array([
        x[1],
        -2 * zeta * omega0 * x[1] - omega0*omega0 * x[0] + u[0]
    ])


def secord_output(t, x, u, params={}):
    """Second order system dynamics output"""
    return np.array([x[0]])


def test_interconnect_unused_input():
    # test that warnings about unused inputs are reported, or not,
    # as required
    g = ct.LinearIOSystem(ct.ss(-1,1,1,0),
                          inputs=['u'],
                          outputs=['y'],
                          name='g')

    s = ct.summing_junction(inputs=['r','-y','-n'],
                            outputs=['e'],
                            name='s')

    k = ct.LinearIOSystem(ct.ss(0,10,2,0),
                          inputs=['e'],
                          outputs=['u'],
                          name='k')

    with pytest.warns(
            UserWarning, match=r"Unused input\(s\) in InterconnectedSystem"):
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'])

    with warnings.catch_warnings():
        # no warning if output explicitly ignored, various argument forms
        warnings.simplefilter("error")
        # strip out matrix warnings
        warnings.filterwarnings("ignore", "the matrix subclass",
                                category=PendingDeprecationWarning)
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_inputs=['n'])

        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_inputs=['s.n'])

        # no warning if auto-connect disabled
        h = ct.interconnect([g,s,k],
                            connections=False)


    # warn if explicity ignored input in fact used
    with pytest.warns(
            UserWarning,
            match=r"Input\(s\) specified as ignored is \(are\) used:") \
            as record:
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_inputs=['u','n'])

    with pytest.warns(
            UserWarning,
            match=r"Input\(s\) specified as ignored is \(are\) used:") \
            as record:
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_inputs=['k.e','n'])

    # error if ignored signal doesn't exist
    with pytest.raises(ValueError):
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_inputs=['v'])


def test_interconnect_unused_output():
    # test that warnings about ignored outputs are reported, or not,
    # as required
    g = ct.LinearIOSystem(ct.ss(-1,1,[[1],[-1]],[[0],[1]]),
                          inputs=['u'],
                          outputs=['y','dy'],
                          name='g')

    s = ct.summing_junction(inputs=['r','-y'],
                            outputs=['e'],
                            name='s')

    k = ct.LinearIOSystem(ct.ss(0,10,2,0),
                          inputs=['e'],
                          outputs=['u'],
                          name='k')

    with pytest.warns(
            UserWarning,
            match=r"Unused output\(s\) in InterconnectedSystem:") as record:
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'])


    # no warning if output explicitly ignored
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        # strip out matrix warnings
        warnings.filterwarnings("ignore", "the matrix subclass",
                                category=PendingDeprecationWarning)
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_outputs=['dy'])

        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_outputs=['g.dy'])

        # no warning if auto-connect disabled
        h = ct.interconnect([g,s,k],
                            connections=False)

    # warn if explicity ignored output in fact used
    with pytest.warns(
            UserWarning,
            match=r"Output\(s\) specified as ignored is \(are\) used:"):
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_outputs=['dy','u'])

    with pytest.warns(
            UserWarning,
            match=r"Output\(s\) specified as ignored is \(are\) used:"):
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_outputs=['dy', ('k.u')])

    # error if ignored signal doesn't exist
    with pytest.raises(ValueError):
        h = ct.interconnect([g,s,k],
                            inputs=['r'],
                            outputs=['y'],
                            ignore_outputs=['v'])


def test_input_output_broadcasting():
    # Create a system, time vector, and noisy input
    sys = ct.rss(6, 2, 3)
    T = np.linspace(0, 10, 10)
    U = np.zeros((sys.ninputs, T.size))
    U[0, :] = np.sin(T)
    U[1, :] = np.zeros_like(U[1, :])
    U[2, :] = np.ones_like(U[2, :])
    X0 = np.array([1, 2])
    P0 = np.array([[3.11, 3.12], [3.21, 3.3]])

    # Simulate the system with nominal input to establish baseline
    resp_base = ct.input_output_response(
        sys, T, U, np.hstack([X0, P0.reshape(-1)]))

    # Split up the inputs into two pieces
    resp_inp1 = ct.input_output_response(sys, T, [U[:1], U[1:]], [X0, P0])
    np.testing.assert_equal(resp_base.states, resp_inp1.states)

    # Specify two of the inputs as constants
    resp_inp2 = ct.input_output_response(sys, T, [U[0], 0, 1], [X0, P0])
    np.testing.assert_equal(resp_base.states, resp_inp2.states)

    # Specify two of the inputs as constant vector
    resp_inp3 = ct.input_output_response(sys, T, [U[0], [0, 1]], [X0, P0])
    np.testing.assert_equal(resp_base.states, resp_inp3.states)

    # Specify only some of the initial conditions
    resp_init = ct.input_output_response(sys, T, [U[0], [0, 1]], [X0, 0])
    resp_cov0 = ct.input_output_response(sys, T, U, [X0, P0 * 0])
    np.testing.assert_equal(resp_cov0.states, resp_init.states)

    # Specify only some of the initial conditions
    with pytest.warns(UserWarning, match="initial state too short; padding"):
        resp_short = ct.input_output_response(sys, T, [U[0], [0, 1]], [X0, 1])

    # Make sure that inconsistent settings don't work
    with pytest.raises(ValueError, match="inconsistent"):
        resp_bad = ct.input_output_response(
            sys, T, (U[0, :], U[:2, :-1]), [X0, P0])

@pytest.mark.parametrize("nstates, ninputs, noutputs", [
    [2, 1, 1],
    [4, 2, 3],
    [0, 1, 1],                  # static function
    [0, 3, 2],                  # static function
])
def test_nonuniform_timepts(nstates, noutputs, ninputs):
    """Test non-uniform time points for simulations"""
    if nstates:
        sys = ct.rss(nstates, noutputs, ninputs)
    else:
        sys = ct.ss(
            [], np.zeros((0, ninputs)), np.zeros((noutputs, 0)),
            np.random.rand(noutputs, ninputs))

    # Start with a uniform set of times
    unifpts = [0, 1, 2, 3, 4,  5,  6,  7,  8,  9, 10]
    uniform = np.outer(
        np.ones(ninputs), [1, 2, 3, 2, 1, -1, -3, -5, -7, -3,  1])
    t_unif, y_unif = ct.input_output_response(
        sys, unifpts, uniform, squeeze=False)

    # Create a non-uniform set of inputs
    noufpts = [0, 2, 4,  8, 10]
    nonunif = np.outer(np.ones(ninputs), [1, 3, 1, -7,  1])
    t_nouf, y_nouf = ct.input_output_response(
        sys, noufpts, nonunif, squeeze=False)

    # Make sure the outputs agree at common times
    np.testing.assert_almost_equal(y_unif[:, noufpts], y_nouf, decimal=6)

    # Resimulate using a new set of evaluation points
    t_even, y_even = ct.input_output_response(
        sys, noufpts, nonunif, t_eval=unifpts, squeeze=False)
    np.testing.assert_almost_equal(y_unif, y_even, decimal=6)


def test_ss_nonlinear():
    """Test ss() for creating nonlinear systems"""
    secord = ct.ss(secord_update, secord_output, inputs='u', outputs='y',
                   states = ['x1', 'x2'], name='secord')
    assert secord.name == 'secord'
    assert secord.input_labels == ['u']
    assert secord.output_labels == ['y']
    assert secord.state_labels == ['x1', 'x2']

    # Make sure we get the same answer for simulations
    T = np.linspace(0, 10, 100)
    U = np.sin(T)
    X0 = np.array([1, -1])
    secord_nlio = ct.NonlinearIOSystem(
        secord_update, secord_output, inputs=1, outputs=1, states=2)
    ss_response = ct.input_output_response(secord, T, U, X0)
    io_response = ct.input_output_response(secord_nlio, T, U, X0)
    np.testing.assert_almost_equal(ss_response.time, io_response.time)
    np.testing.assert_almost_equal(ss_response.inputs, io_response.inputs)
    np.testing.assert_almost_equal(ss_response.outputs, io_response.outputs)

    # Make sure that optional keywords are allowed
    secord = ct.ss(secord_update, secord_output, dt=True)
    assert ct.isdtime(secord)

    # Make sure that state space keywords are flagged
    with pytest.raises(TypeError, match="unrecognized keyword"):
        ct.ss(secord_update, remove_useless_states=True)


def test_rss():
    # Basic call, with no arguments
    sys = ct.rss()
    assert sys.ninputs == 1
    assert sys.noutputs == 1
    assert sys.nstates == 1
    assert sys.dt == 0
    assert np.all(np.real(sys.poles()) < 0)

    # Set the timebase explicitly
    sys = ct.rss(inputs=2, outputs=3, states=4, dt=None, name='sys')
    assert sys.name == 'sys'
    assert sys.ninputs == 2
    assert sys.noutputs == 3
    assert sys.nstates == 4
    assert sys.dt == None
    assert np.all(np.real(sys.poles()) < 0)

    # Discrete time
    sys = ct.rss(inputs=['a', 'b'], outputs=1, states=1, dt=True)
    assert sys.ninputs == 2
    assert sys.input_labels == ['a', 'b']
    assert sys.noutputs == 1
    assert sys.nstates == 1
    assert sys.dt == True
    assert np.all(np.abs(sys.poles()) < 1)

    # Call drss directly
    sys = ct.drss(inputs=['a', 'b'], outputs=1, states=1, dt=True)
    assert sys.ninputs == 2
    assert sys.input_labels == ['a', 'b']
    assert sys.noutputs == 1
    assert sys.nstates == 1
    assert sys.dt == True
    assert np.all(np.abs(sys.poles()) < 1)

    with pytest.raises(ValueError, match="continuous timebase"):
        sys = ct.drss(2, 1, 1, dt=0)

    with pytest.warns(UserWarning, match="may be interpreted as continuous"):
        sys = ct.drss(2, 1, 1, dt=None)
        assert np.all(np.abs(sys.poles()) < 1)


def eqpt_rhs(t, x, u, params):
    return np.array([x[0]/2 + u[0], x[0] - x[1]**2 + u[1], x[1] - x[2]])

def eqpt_out(t, x, u, params):
    return np.array([x[0], x[1] + u[1]])

@pytest.mark.parametrize(
    "x0, ix, u0, iu, y0, iy, dx0, idx, dt, x_expect, u_expect", [
        # Equilibrium points with input given
        (0, None, 0, None, None, None, None, None, 0, [0, 0, 0], [0, 0]),
        (0, None, 0, None, None, None, None, None, None, [0, 0, 0], [0, 0]),
        ([0.9, 0.9, 0.9], None, [-1, 0], None, None, None, None, None, 0,
         [2, sqrt(2), sqrt(2)], [-1, 0]),
        ([0.9, -0.9, 0.9], None, [-1, 0], None, None, None, None, None, 0,
         [2, -sqrt(2), -sqrt(2)], [-1, 0]),     # same input, different eqpt
        (0, None, 0, None, None, None, None, None, 1, [0, 0, 0], [0, 0]), #DT
        (0, None, [-1, 0], None, None, None, None, None, 1, None, None),  #DT
        ([0, -0.1, 0], None, [0, -0.25], None, None, None, None, None, 1, #DT
         [0, -0.5, -0.25], [0, -0.25]),

        # Equilibrium points with output given
        ([0.9, 0.9, 0.9], None, [-0.9, 0], None, [2, sqrt(2)], None, None,
         None, 0, [2, sqrt(2), sqrt(2)], [-1, 0]),
        (0, None, [0, -0.25], None, [0, -0.75], None, None, None, 1,      #DT
         [0, -0.5, -0.25], [0, -0.25]),

        # Equilibrium points with mixture of inputs and outputs given
        ([0.9, 0.9, 0.9], None, [-1, 0], [0], [2, sqrt(2)], [1], None,
         None, 0, [2, sqrt(2), sqrt(2)], [-1, 0]),
        (0, None, [0, -0.22], [0], [0, -0.75], [1], None, None, 1,        #DT
         [0, -0.5, -0.25], [0, -0.25]),
    ])

def test_find_eqpt(x0, ix, u0, iu, y0, iy, dx0, idx, dt, x_expect, u_expect):
    sys = ct.NonlinearIOSystem(
        eqpt_rhs, eqpt_out, dt=dt, states=3, inputs=2, outputs=2)

    xeq, ueq = ct.find_eqpt(
        sys, x0, u0, y0, ix=ix, iu=iu, iy=iy, dx0=dx0, idx=idx)

    # If no equilibrium points, skip remaining tests
    if x_expect is None:
        assert xeq is None
        assert ueq is None
        return

    # Make sure we are at an appropriate equilibrium point
    if dt is None or dt == 0:
        # Continuous time system
        np.testing.assert_allclose(eqpt_rhs(0, xeq, ueq, {}), 0, atol=1e-6)
        if y0 is not None:
            y0 = np.array(y0)
            iy = np.s_[:] if iy is None else np.array(iy)
            np.testing.assert_allclose(
                eqpt_out(0, xeq, ueq, {})[iy], y0[iy], atol=1e-6)

    else:
        # Discrete time system
        np.testing.assert_allclose(eqpt_rhs(0, xeq, ueq, {}), xeq, atol=1e-6)
        if y0 is not None:
            y0 = np.array(y0)
            iy = np.s_[:] if iy is None else np.array(iy)
            np.testing.assert_allclose(
                eqpt_out(0, xeq, ueq, {})[iy], y0[iy], atol=1e-6)

    # Check that we got the expected result as well
    np.testing.assert_allclose(np.array(xeq), x_expect, atol=1e-6)
    np.testing.assert_allclose(np.array(ueq), u_expect, atol=1e-6)



"""lti.py

The lti module contains the LTI parent class to the child classes StateSpace
and TransferFunction.  It is designed for use in the python-control library.

Routines in this module:

LTI.__init__
isdtime()
isctime()
timebase()
common_timebase()
"""

import numpy as np

from numpy import absolute, real, angle, abs
from warnings import warn
from . import config
from .namedio import NamedIOSystem, isdtime

__all__ = ['poles', 'zeros', 'damp', 'evalfr', 'frequency_response',
           'freqresp', 'dcgain', 'pole', 'zero']


class LTI(NamedIOSystem):
    """LTI is a parent class to linear time-invariant (LTI) system objects.

    LTI is the parent to the StateSpace and TransferFunction child classes. It
    contains the number of inputs and outputs, and the timebase (dt) for the
    system.  This function is not generally called directly by the user.

    The timebase for the system, dt, is used to specify whether the system
    is operating in continuous or discrete time. It can have the following
    values:

      * dt = None       No timebase specified
      * dt = 0          Continuous time system
      * dt > 0          Discrete time system with sampling time dt
      * dt = True       Discrete time system with unspecified sampling time

    When two LTI systems are combined, their timebases much match.  A system
    with timebase None can be combined with a system having a specified
    timebase, and the result will have the timebase of the latter system.

    Note: dt processing has been moved to the NamedIOSystem class.

    """

    def __init__(self, inputs=1, outputs=1, states=None, name=None, **kwargs):
        """Assign the LTI object's numbers of inputs and ouputs."""
        super().__init__(
            name=name, inputs=inputs, outputs=outputs, states=states, **kwargs)

    #
    # Getter and setter functions for legacy state attributes
    #
    # For this iteration, generate a deprecation warning whenever the
    # getter/setter is called.  For a future iteration, turn it into a
    # future warning, so that users will see it.
    #

    def _get_inputs(self):
        warn("The LTI `inputs` attribute will be deprecated in a future "
             "release.  Use `ninputs` instead.",
             DeprecationWarning, stacklevel=2)
        return self.ninputs

    def _set_inputs(self, value):
        warn("The LTI `inputs` attribute will be deprecated in a future "
             "release.  Use `ninputs` instead.",
             DeprecationWarning, stacklevel=2)
        self.ninputs = value

    #: Deprecated
    inputs = property(
        _get_inputs, _set_inputs, doc="""
        Deprecated attribute; use :attr:`ninputs` instead.

        The ``inputs`` attribute was used to store the number of system
        inputs.  It is no longer used.  If you need access to the number
        of inputs for an LTI system, use :attr:`ninputs`.
        """)

    def _get_outputs(self):
        warn("The LTI `outputs` attribute will be deprecated in a future "
             "release.  Use `noutputs` instead.",
             DeprecationWarning, stacklevel=2)
        return self.noutputs

    def _set_outputs(self, value):
        warn("The LTI `outputs` attribute will be deprecated in a future "
             "release.  Use `noutputs` instead.",
             DeprecationWarning, stacklevel=2)
        self.noutputs = value

    #: Deprecated
    outputs = property(
        _get_outputs, _set_outputs, doc="""
        Deprecated attribute; use :attr:`noutputs` instead.

        The ``outputs`` attribute was used to store the number of system
        outputs.  It is no longer used.  If you need access to the number of
        outputs for an LTI system, use :attr:`noutputs`.
        """)

    def damp(self):
        '''Natural frequency, damping ratio of system poles

        Returns
        -------
        wn : array
            Natural frequencies for each system pole
        zeta : array
            Damping ratio for each system pole
        poles : array
            Array of system poles
        '''
        poles = self.poles()

        if self.isdtime(strict=True):
            splane_poles = np.log(poles.astype(complex))/self.dt
        else:
            splane_poles = poles
        wn = absolute(splane_poles)
        Z = -real(splane_poles)/wn
        return wn, Z, poles

    def frequency_response(self, omega, squeeze=None):
        """Evaluate the linear time-invariant system at an array of angular
        frequencies.

        Reports the frequency response of the system,

             G(j*omega) = mag * exp(j*phase)

        for continuous time systems. For discrete time systems, the response
        is evaluated around the unit circle such that

             G(exp(j*omega*dt)) = mag * exp(j*phase).

        In general the system may be multiple input, multiple output (MIMO),
        where `m = self.ninputs` number of inputs and `p = self.noutputs`
        number of outputs.

        Parameters
        ----------
        omega : float or 1D array_like
            A list, tuple, array, or scalar value of frequencies in
            radians/sec at which the system will be evaluated.
        squeeze : bool, optional
            If squeeze=True, remove single-dimensional entries from the shape
            of the output even if the system is not SISO. If squeeze=False,
            keep all indices (output, input and, if omega is array_like,
            frequency) even if the system is SISO. The default value can be
            set using config.defaults['control.squeeze_frequency_response'].

        Returns
        -------
        response : :class:`FrequencyReponseData`
            Frequency response data object representing the frequency
            response.  This object can be assigned to a tuple using

                mag, phase, omega = response

            where ``mag`` is the magnitude (absolute value, not dB or
            log10) of the system frequency response, ``phase`` is the wrapped
            phase in radians of the system frequency response, and ``omega``
            is the (sorted) frequencies at which the response was evaluated.
            If the system is SISO and squeeze is not True, ``magnitude`` and
            ``phase`` are 1D, indexed by frequency.  If the system is not SISO
            or squeeze is False, the array is 3D, indexed by the output,
            input, and frequency.  If ``squeeze`` is True then
            single-dimensional axes are removed.

        """
        omega = np.sort(np.array(omega, ndmin=1))
        if self.isdtime(strict=True):
            # Convert the frequency to discrete time
            if np.any(omega * self.dt > np.pi):
                warn("__call__: evaluation above Nyquist frequency")
            s = np.exp(1j * omega * self.dt)
        else:
            s = 1j * omega

        # Return the data as a frequency response data object
        from .frdata import FrequencyResponseData
        response = self.__call__(s)
        return FrequencyResponseData(
            response, omega, return_magphase=True, squeeze=squeeze)

    def dcgain(self):
        """Return the zero-frequency gain"""
        raise NotImplementedError("dcgain not implemented for %s objects" %
                                  str(self.__class__))

    def _dcgain(self, warn_infinite):
        zeroresp = self(0 if self.isctime() else 1,
                        warn_infinite=warn_infinite)
        if np.all(np.logical_or(np.isreal(zeroresp), np.isnan(zeroresp.imag))):
            return zeroresp.real
        else:
            return zeroresp

    def ispassive(self):
        # importing here prevents circular dependancy
        from control.passivity import ispassive
        return ispassive(self)

    #
    # Deprecated functions
    #

    def pole(self):
        warn("pole() will be deprecated; use poles()",
             PendingDeprecationWarning)
        return self.poles()

    def zero(self):
        warn("zero() will be deprecated; use zeros()",
             PendingDeprecationWarning)
        return self.zeros()


def poles(sys):
    """
    Compute system poles.

    Parameters
    ----------
    sys: StateSpace or TransferFunction
        Linear system

    Returns
    -------
    poles: ndarray
        Array that contains the system's poles.

    See Also
    --------
    zeros
    TransferFunction.poles
    StateSpace.poles

    """

    return sys.poles()


def pole(sys):
    warn("pole() will be deprecated; use poles()", PendingDeprecationWarning)
    return poles(sys)


def zeros(sys):
    """
    Compute system zeros.

    Parameters
    ----------
    sys: StateSpace or TransferFunction
        Linear system

    Returns
    -------
    zeros: ndarray
        Array that contains the system's zeros.

    See Also
    --------
    poles
    StateSpace.zeros
    TransferFunction.zeros

    """

    return sys.zeros()


def zero(sys):
    warn("zero() will be deprecated; use zeros()", PendingDeprecationWarning)
    return zeros(sys)


def damp(sys, doprint=True):
    """
    Compute natural frequency, damping ratio, and poles of a system

    The function takes 1 or 2 parameters

    Parameters
    ----------
    sys: LTI (StateSpace or TransferFunction)
        A linear system object
    doprint:
        if true, print table with values

    Returns
    -------
    wn: array
        Natural frequencies of the poles
    damping: array
        Damping values
    poles: array
        Pole locations

    See Also
    --------
    pole

    Notes
    -----
    If the system is continuous,
        wn = abs(poles)
        Z  = -real(poles)/poles.

    If the system is discrete, the discrete poles are mapped to their
    equivalent location in the s-plane via

        s = log10(poles)/dt

    and

        wn = abs(s)
        Z = -real(s)/wn.

    """
    wn, damping, poles = sys.damp()
    if doprint:
        print('_____Eigenvalue______ Damping___ Frequency_')
        for p, d, w in zip(poles, damping, wn):
            if abs(p.imag) < 1e-12:
                print("%10.4g            %10.4g %10.4g" %
                      (p.real, 1.0, -p.real))
            else:
                print("%10.4g%+10.4gj %10.4g %10.4g" %
                      (p.real, p.imag, d, w))
    return wn, damping, poles


def evalfr(sys, x, squeeze=None):
    """Evaluate the transfer function of an LTI system for complex frequency x.

    Returns the complex frequency response `sys(x)` where `x` is `s` for
    continuous-time systems and `z` for discrete-time systems, with
    `m = sys.ninputs` number of inputs and `p = sys.noutputs` number of
    outputs.

    To evaluate at a frequency omega in radians per second, enter
    ``x = omega * 1j`` for continuous-time systems, or
    ``x = exp(1j * omega * dt)`` for discrete-time systems, or use
    ``freqresp(sys, omega)``.

    Parameters
    ----------
    sys: StateSpace or TransferFunction
        Linear system
    x : complex scalar or 1D array_like
        Complex frequency(s)
    squeeze : bool, optional (default=True)
        If squeeze=True, remove single-dimensional entries from the shape of
        the output even if the system is not SISO. If squeeze=False, keep all
        indices (output, input and, if omega is array_like, frequency) even if
        the system is SISO. The default value can be set using
        config.defaults['control.squeeze_frequency_response'].

    Returns
    -------
    fresp : complex ndarray
        The frequency response of the system.  If the system is SISO and
        squeeze is not True, the shape of the array matches the shape of
        omega.  If the system is not SISO or squeeze is False, the first two
        dimensions of the array are indices for the output and input and the
        remaining dimensions match omega.  If ``squeeze`` is True then
        single-dimensional axes are removed.

    See Also
    --------
    freqresp
    bode

    Notes
    -----
    This function is a wrapper for :meth:`StateSpace.__call__` and
    :meth:`TransferFunction.__call__`.

    Examples
    --------
    >>> sys = ss("1. -2; 3. -4", "5.; 7", "6. 8", "9.")
    >>> evalfr(sys, 1j)
    array([[ 44.8-21.4j]])
    >>> # This is the transfer function matrix evaluated at s = i.

    .. todo:: Add example with MIMO system

    """
    return sys.__call__(x, squeeze=squeeze)


def frequency_response(sys, omega, squeeze=None):
    """Frequency response of an LTI system at multiple angular frequencies.

    In general the system may be multiple input, multiple output (MIMO), where
    `m = sys.ninputs` number of inputs and `p = sys.noutputs` number of
    outputs.

    Parameters
    ----------
    sys: StateSpace or TransferFunction
        Linear system
    omega : float or 1D array_like
        A list of frequencies in radians/sec at which the system should be
        evaluated. The list can be either a python list or a numpy array
        and will be sorted before evaluation.
    squeeze : bool, optional
        If squeeze=True, remove single-dimensional entries from the shape of
        the output even if the system is not SISO. If squeeze=False, keep all
        indices (output, input and, if omega is array_like, frequency) even if
        the system is SISO. The default value can be set using
        config.defaults['control.squeeze_frequency_response'].

    Returns
    -------
    response : FrequencyResponseData
        Frequency response data object representing the frequency response.
        This object can be assigned to a tuple using

            mag, phase, omega = response

        where ``mag`` is the magnitude (absolute value, not dB or log10) of
        the system frequency response, ``phase`` is the wrapped phase in
        radians of the system frequency response, and ``omega`` is the
        (sorted) frequencies at which the response was evaluated.  If the
        system is SISO and squeeze is not True, ``magnitude`` and ``phase``
        are 1D, indexed by frequency.  If the system is not SISO or squeeze
        is False, the array is 3D, indexed by the output, input, and
        frequency.  If ``squeeze`` is True then single-dimensional axes are
        removed.

    See Also
    --------
    evalfr
    bode

    Notes
    -----
    This function is a wrapper for :meth:`StateSpace.frequency_response` and
    :meth:`TransferFunction.frequency_response`.

    Examples
    --------
    >>> sys = ss("1. -2; 3. -4", "5.; 7", "6. 8", "9.")
    >>> mag, phase, omega = freqresp(sys, [0.1, 1., 10.])
    >>> mag
    array([[[ 58.8576682 ,  49.64876635,  13.40825927]]])
    >>> phase
    array([[[-0.05408304, -0.44563154, -0.66837155]]])

    .. todo::
        Add example with MIMO system

        #>>> sys = rss(3, 2, 2)
        #>>> mag, phase, omega = freqresp(sys, [0.1, 1., 10.])
        #>>> mag[0, 1, :]
        #array([ 55.43747231,  42.47766549,   1.97225895])
        #>>> phase[1, 0, :]
        #array([-0.12611087, -1.14294316,  2.5764547 ])
        #>>> # This is the magnitude of the frequency response from the 2nd
        #>>> # input to the 1st output, and the phase (in radians) of the
        #>>> # frequency response from the 1st input to the 2nd output, for
        #>>> # s = 0.1i, i, 10i.

    """
    return sys.frequency_response(omega, squeeze=squeeze)


# Alternative name (legacy)
freqresp = frequency_response


def dcgain(sys):
    """Return the zero-frequency (or DC) gain of the given system

    Returns
    -------
    gain : ndarray
        The zero-frequency gain, or (inf + nanj) if the system has a pole at
        the origin, (nan + nanj) if there is a pole/zero cancellation at the
        origin.

    """
    return sys.dcgain()


# Process frequency responses in a uniform way
def _process_frequency_response(sys, omega, out, squeeze=None):
    # Set value of squeeze argument if not set
    if squeeze is None:
        squeeze = config.defaults['control.squeeze_frequency_response']

    if np.asarray(omega).ndim < 1:
        # received a scalar x, squeeze down the array along last dim
        out = np.squeeze(out, axis=2)

    #
    # Get rid of unneeded dimensions
    #
    # There are three possible values for the squeeze keyword at this point:
    #
    #   squeeze=None: squeeze input/output axes iff SISO
    #   squeeze=True: squeeze all single dimensional axes (ala numpy)
    #   squeeze-False: don't squeeze any axes
    #
    if squeeze is True:
        # Squeeze everything that we can if that's what the user wants
        return np.squeeze(out)
    elif squeeze is None and sys.issiso():
        # SISO system output squeezed unless explicitly specified otherwise
        return out[0][0]
    elif squeeze is False or squeeze is None:
        return out
    else:
        raise ValueError("unknown squeeze value")
